/* CommonStuff.c */

#define compilingcommonstuff

#include "commonstuff.h"
#include "crc32.h"


MyBoolean   QuitFlag = 0;


void        ErrorEncode(TransRec* Packet)
  {
    ulong     CRC;
    short     NumBytes;

    NumBytes = Packet->Length;
    CRC = UpdateCRC(Packet->Data,NumBytes);
    Packet->Data[NumBytes + 0] = CRC >> 24;
    Packet->Data[NumBytes + 1] = CRC >> 16;
    Packet->Data[NumBytes + 2] = CRC >> 8;
    Packet->Data[NumBytes + 3] = CRC;
    Packet->Length += 4;
  }


/* Returns TRUE if data is correct or FALSE if it is bad. */
MyBoolean   ErrorDecode(char* Data, short* Length)
  {
    ulong     CRC;
    short     NumDataBytes;

    (*Length) -= 4;
    NumDataBytes = (*Length);
    CRC = UpdateCRC(Data,NumDataBytes);
    return ((NumDataBytes >= 0) && (0 == (0x000000ff & (
      (Data[NumDataBytes + 0] ^ (CRC >> 24))
      | (Data[NumDataBytes + 1] ^ (CRC >> 16))
      | (Data[NumDataBytes + 2] ^ (CRC >> 8))
      | (Data[NumDataBytes + 3] ^ CRC)))));
  }


void        TransmitEncode(TransRec* Packet, short NumBits)
  {
    switch (NumBits)
      {
        case 8:
          TransmitEncode8(Packet);
          break;
        case 7:
          TransmitEncode7(Packet);
          break;
      }
  }


/* you have to pass a packet in, so when the packet is being encoded, */
/* ALL of the available buffer space may be used */
void        TransmitEncode8(TransRec* Packet)
  {
    uchar     Mask;
    short     Index;
    short     DataScan;
    short     NumResolutionBytes;

    /* setting up encoding routine. */
    Index = MaxTotalPacketSize - 1; /* start putting bit array at top of block */
    NumResolutionBytes = 1;
    Mask = 0x40; /* we don't use the 8th bit */
    Packet->Data[Index] = 0;  /* zero the first entry */

    /* stepping through all data to generate masks & data */
    DataScan = 0;
    while (DataScan < Packet->Length)
      {
        uchar     Temp;

        Temp = Packet->Data[DataScan]; /* get a byte */
        if ((Temp < 32) || (Temp >= 224))
          {
            /* there is a possibility of ambiguity, so encode a bit */
            if (Temp >= 224)
              {
                /* if it is a high value, encode a 1, leave it at 0 */
                Packet->Data[Index] |= Mask;
                Temp -= 224;  /* take modulo 224 of temp */
              }
            Mask = Mask >> 1;
            if (Mask == 0)
              {
                /* step mask to next byte */
                Mask = 0x40;
                Index -= 1; /* remember, we descend */
                NumResolutionBytes += 1;
                Packet->Data[Index] = 0; /* zero the next entry */
              }
          }
        Packet->Data[DataScan] = Temp + 32;  /* store byte into output stream. */
        DataScan += 1;
      }
    /* At exit, if Mask = 0x40, then Index points to the byte after the last, otherwise */
    /* it points to the last byte of data. */
    if (Mask == 0x40)
      {
        Index += 1;  /* don't include last byte if it wasn't actually used. */
        NumResolutionBytes -= 1;
      }
    /* now we have to copy the list of bits down on top of the data */
    for (DataScan = 0; DataScan < NumResolutionBytes; DataScan += 1)
      {
        Packet->Data[Packet->Length] = Packet->Data[Index] + 32;
        Index += 1;
        Packet->Length += 1;
      }
    Packet->Data[Packet->Length] = 0x0d;  /* packet terminator character */
    Packet->Length += 1;
  }


/* TransmitEncode7 does something akin to what TransmitEncode8 does, */
/* but first it streches the 8-bit bytes into 7-bit bytes. */
/* you have to pass a packet in, so when the packet is being encoded, */
/* ALL of the available buffer space may be used */
void        TransmitEncode7(TransRec* Packet)
  {
    uchar     Mask;
    short     Index;
    short     DataScan;
    short     NumResolutionBytes;
    short     RawIndex;
    short     ExtraBits;
    char      ShadowChar;

    /* streching... */
    /* 0 bytes == 0 bits == 0.00 bytes == 0 bytes */
    /* 1 byte == 8 bits == 1.14 bytes == 2 bytes */
    /* 2 bytes == 16 bits == 2.29 bytes == 3 bytes */
    /* 3 bytes == 24 bits == 3.43 bytes == 4 bytes */
    /* 4 bytes == 32 bits == 4.57 bytes == 5 bytes */
    /* 5 bytes == 40 bits == 5.71 bytes == 6 bytes */
    /* 6 bytes == 48 bits == 6.86 bytes == 7 bytes */
    /* 7 bytes == 56 bits == 8.00 bytes == 8 bytes */
    /* 8 bytes == 64 bits == 9.14 bytes == 10 bytes */
    /* The pattern:  For each block of 7 bytes, add another byte. */
    /* if there are any left over, add yet another byte */

    /* first, copy all data to top of packet buffer */
    RawIndex = MaxTotalPacketSize - Packet->Length;
    for (Index = Packet->Length - 1; Index >= 0; Index -= 1)
      {
        Packet->Data[RawIndex + Index] = Packet->Data[Index];
      }

    /* now, we  s t r e c h */
    DataScan = 0;
    ExtraBits = 0;
    Mask = 0;
    for (Index = RawIndex; Index < MaxTotalPacketSize; Index += 1)
      {
        /* move low 7 bits over unchanged, mask out high bit */
        ShadowChar = Packet->Data[Index];
        Packet->Data[DataScan] = ShadowChar & 0x007f;
        Mask = Mask << 1; /* open a hole on the bottom of mask */
        /* notice how this is shifted left.  That means the oldest bit */
        /* will be at the top.  Therefore, the first byte's bit is at */
        /* the top and the last byte's bit is at the bottom */
        if ((ShadowChar & 0x0080) != 0)
          {
            /* if high bit is set, then set Mask bit, else leave it clear */
            Mask |= 1;
          }
        ExtraBits += 1;
        DataScan += 1;
        if (ExtraBits == 6)
          {
            /* if we have a whole byte worth of extra-bits, then we store them too */
            Packet->Data[DataScan] = Mask & 0x003f; /* 6 significant bits */
            DataScan += 1;
            ExtraBits = 0;
            Mask = 0;
          }
      }
    if (ExtraBits != 0)
      {
        /* if there were some bits left over at the end, then we have to */
        /* store them too. */
        Packet->Data[DataScan] = Mask;
        DataScan += 1;
      }
    Packet->Length = DataScan; /* this is now the new effective length */

    /* setting up encoding routine. */
    Index = MaxTotalPacketSize - 1; /* start putting bit array at top of block */
    NumResolutionBytes = 1;
    Mask = 0x20; /* we don't use the 8th or 7th bits */
    Packet->Data[Index] = 0;  /* zero the first entry */

    /* stepping through all data to generate masks & data */
    DataScan = 0;
    while (DataScan < Packet->Length)
      {
        uchar     Temp;

        Temp = Packet->Data[DataScan]; /* get a byte */
        if ((Temp < 32) || (Temp >= 96))
          {
            /* there is a possibility of ambiguity, so encode a bit */
            if (Temp >= 96)
              {
                /* if it is a high value, encode a 1, leave it at 0 */
                Packet->Data[Index] |= Mask;
                Temp -= 96;  /* take modulo 96 of temp */
              }
            Mask = Mask >> 1;
            if (Mask == 0)
              {
                /* step mask to next byte */
                Mask = 0x20;
                Index -= 1; /* remember, we descend */
                NumResolutionBytes += 1;
                Packet->Data[Index] = 0; /* zero the next entry */
              }
          }
        Packet->Data[DataScan] = Temp + 32;  /* store byte into output stream. */
        DataScan += 1;
      }
    /* At exit, if Mask = 0x20, then Index points to the byte after the last, otherwise */
    /* it points to the last byte of data. */
    if (Mask == 0x20)
      {
        Index += 1;  /* don't include last byte if it wasn't actually used. */
        NumResolutionBytes -= 1;
      }
    /* now we have to copy the list of bits down on top of the data */
    for (DataScan = 0; DataScan < NumResolutionBytes; DataScan += 1)
      {
        Packet->Data[Packet->Length] = Packet->Data[Index] + 32;
        Index += 1;
        Packet->Length += 1;
      }
    Packet->Data[Packet->Length] = 0x0d;  /* packet terminator character */
    Packet->Length += 1;
  }


/* this decodes data that has been received from the channel using the reverse of */
/* the method mentioned above.  The original handle is released.  There should NOT */
/* be a carriage return on the end at this point. */
MyBoolean   TransmitDecode(char* Data, short* Length, short NumBits)
  {
    switch (NumBits)
      {
        case 8:
          return TransmitDecode8(Data,Length);
        case 7:
          return TransmitDecode7(Data,Length);
        default:
          return False;
      }
  }


/* The pointer passed in is right from the data buffer, so this routine can */
/* not use any more than 'Length' bytes (i.e. modify in place, don't overflow) */
MyBoolean   TransmitDecode8(char* Data, short* Length)
  {
    short     DataIndex;
    short     BitCheckIndex;
    char      Mask;

    /* setting up the routine. */
    Mask = 0x40;
    DataIndex = 0; /* data bytes start at beginning and go forwards. */
    BitCheckIndex = (*Length) - 1; /* ambiguity resolution bytes start at end & go backwards. */

    /* scanning data and resolving it. */
    /* we go until DataIndex >= BitCheckIndex.  If all goes well, they should */
    /* end up being equal.  In the rare chance they arn't, then something is */
    /* missing in the data block, so we return False indicating an error. */
    while ((DataIndex < BitCheckIndex)
      || ((DataIndex == BitCheckIndex) && (Mask == 0x40)))
      {
        uchar     Temp;

        Temp = Data[DataIndex] - 32;
        if (Temp >= 224)
          {
            return False;  /* shouldn't ever happen! */
          }
        if (Temp < 32)
          {
            if (((Data[BitCheckIndex] - 32) & Mask) != 0)
              {
                Temp += 224; /* add in if bit was set, else leave low */
              }
            Mask = Mask >> 1;
            if (Mask == 0)
              {
                Mask = 0x40;
                BitCheckIndex -= 1;
              }
          }
        Data[DataIndex] = Temp;
        DataIndex += 1;
      }
    /* at exit:  EncodedData[DataIndex] should be FIRST resolution byte. */
    /* It's also the total number of data bytes. */
    /* and EncodedData[BitCheckIndex] should be FIRST resolution byte, OR */
    /* if Mask == 0x40, then it wrapped around falsly, and should be */
    /* LAST data byte. */
    if (Mask == 0x40)
      {
        BitCheckIndex += 1; /* now we fix the wrap-around. */
      }

    (*Length) = DataIndex; /* remote resolution bits from end */
    return (BitCheckIndex == DataIndex);  /* return error status. */
  }


/* this does the same sort of thing as TransmitDecode8, but unstreches */
/* the data after */
/* The pointer passed in is right from the data buffer, so this routine can */
/* not use any more than 'Length' bytes (i.e. modify in place, don't overflow) */
MyBoolean   TransmitDecode7(char* Data, short* Length)
  {
    short     DataIndex;
    short     BitCheckIndex;
    char      Mask;
    short     FinalIndex;
    short     Count;

    /* setting up the routine. */
    Mask = 0x20;
    DataIndex = 0; /* data bytes start at beginning and go forwards. */
    BitCheckIndex = (*Length) - 1; /* ambiguity resolution bytes start at end & go backwards. */

    /* scanning data and resolving it. */
    /* we go until DataIndex >= BitCheckIndex.  If all goes well, they should */
    /* end up being equal.  In the rare chance they arn't, then something is */
    /* missing in the data block, so we return False indicating an error. */
    while ((DataIndex < BitCheckIndex)
      || ((DataIndex == BitCheckIndex) && (Mask == 0x20)))
      {
        uchar     Temp;

        Temp = (Data[DataIndex]) - 32;
        if (Temp >= 96)
          {
            return False;  /* shouldn't ever happen! */
          }
        if (Temp < 32)
          {
            if (((Data[BitCheckIndex] - 32) & Mask) != 0)
              {
                Temp += 96; /* add in if bit was set, else leave low */
              }
            Mask = Mask >> 1;
            if (Mask == 0)
              {
                Mask = 0x20;
                BitCheckIndex -= 1;
              }
          }
        Data[DataIndex] = Temp;
        DataIndex += 1;
      }
    /* at exit:  EncodedData[DataIndex] should be FIRST resolution byte. */
    /* It's also the total number of data bytes. */
    /* and EncodedData[BitCheckIndex] should be FIRST resolution byte, OR */
    /* if Mask == 0x20, then it wrapped around falsly, and should be */
    /* LAST data byte. */
    if (Mask == 0x20)
      {
        BitCheckIndex += 1; /* now we fix the wrap-around. */
      }

    (*Length) = DataIndex; /* remote resolution bits from end */
    if (BitCheckIndex != DataIndex)
      {
        return False;  /* this means the ambiguity resolution failed */
      }

    /* now it's time to unstrech. */
    DataIndex = 0;
    FinalIndex = 0;
    while ((*Length) - DataIndex >= 7)
      {
        /* while there is still some stuff to do */
        /* the bits DataIndex+0 .. DataIndex+5 are all fine, except they are missing */
        /* the 8th bit, which is packed into DataIndex+6.  This makes for a total */
        /* of 7 bytes.  If there are less than that, then the packed bits byte is */
        /* the last one */
        for (Count = 0; Count < 6; Count += 1)
          {
            Data[FinalIndex] = Data[DataIndex + Count];
            /* oldest bit [0] is in position 5.  newest [5] is in position 0 */
            /* thus, shift %00100000 (0x20) right by offset from start. */
            if ((Data[DataIndex + 6] & (0x20 >> Count)) != 0)
              {
                Data[FinalIndex] |= 0x80; /* set the high bit */
              }
            FinalIndex += 1;
          }
        DataIndex += 6 + 1; /* 6 data bytes + the extra bits byte */
      }
    /* this cleans up any non multiple of 6 byte block.  If there was a */
    /* multiple of 6 bytes, (raw, not encoded) then (*Length) - DataIndex - 1 == -1, */
    /* so this loop will not be executed. */
    for (Count = 0; Count < (*Length) - DataIndex - 1; Count += 1)
      {
        Data[FinalIndex] = Data[DataIndex + Count];
        /* oldest bit [0] is in position 5.  newest [5] is in position 0 */
        /* thus, shift %00100000 (0x20) right by offset from start. */
        if ((Data[(*Length) - 1]
          & (0x20 >> (Count + 6 - ((*Length) - DataIndex - 1)))) != 0)
          {
            /* the Count + 6 - ((*Length) - DataIndex - 1) is because */
            /* if there are fewer than 6 bits, they are packed in the low */
            /* end of the byte */
            Data[FinalIndex] |= 0x80; /* set the high bit */
          }
        FinalIndex += 1;
      }
    (*Length) = FinalIndex;
    return True; /* if we got this far, then the data is probably good. */
    /* we could check to see that the last byte didn't have any extraneous */
    /* bits, but they probably wouldn't matter much anyway, and I don't */
    /* feel like doing it. */
  }


/****************************************************************/
/* Getting and Setting and Getting and Setting and Getting TTY Stuff */


void    GetTTYInfo(int TheFile, TTYInfo* TheInfo)
  {
    APRINT(("+GetTTYInfo\n"));
    ioctl(TheFile,TIOCGETC,&(TheInfo->TCHARS));
    ioctl(TheFile,TIOCGETD,&(TheInfo->LDISC));
    ioctl(TheFile,TIOCGETP,&(TheInfo->TTYB));
    ioctl(TheFile,TIOCGLTC,&(TheInfo->LTCHARS));
    ioctl(TheFile,TIOCLGET,&(TheInfo->LMODE));
    TheInfo->fcntlFlags = fcntl(TheFile,F_GETFL,0);
    APRINT(("-GetTTYInfo\n"));
  }


void    SetTTYInfo(int TheFile, TTYInfo* TheInfo)
  {
    APRINT(("+SetTTYInfo\n"));
    ioctl(TheFile,TIOCSETC,&(TheInfo->TCHARS));
    ioctl(TheFile,TIOCSETD,&(TheInfo->LDISC));
    ioctl(TheFile,TIOCSETP,&(TheInfo->TTYB));
    ioctl(TheFile,TIOCSLTC,&(TheInfo->LTCHARS));
    ioctl(TheFile,TIOCLSET,&(TheInfo->LMODE));
    fcntl(TheFile,F_SETFL,TheInfo->fcntlFlags);
/*
    error = ioctl(TheFile,TIOCEXCL,0);
    ERROR(error==-1,APRINT(("SetTTYInfo ioctl for exclusive use error %d\n",errno)));
*/
    APRINT(("-SetTTYInfo\n"));
  }


void    MakeSpiffyTTY(TTYInfo* TheInfo)
  {
    APRINT(("+MakeSpiffyTTY\n"));
    TheInfo->TTYB.sg_flags = RAW | ANYP;
    TheInfo->TCHARS.t_intrc = -1;
    TheInfo->TCHARS.t_quitc = -1;
    TheInfo->TCHARS.t_startc = -1;
    TheInfo->TCHARS.t_stopc = -1;
    TheInfo->LTCHARS.t_suspc = -1;
    TheInfo->LTCHARS.t_dsuspc = -1;
    TheInfo->LTCHARS.t_flushc = -1;
    TheInfo->LTCHARS.t_lnextc = -1;
    TheInfo->fcntlFlags |= O_NDELAY;
    APRINT(("-MakeSpiffyTTY\n"));
  }



#ifdef dBUG

FILE*   bugfile;

void    AuditPrint(char* First, ...)
  {
    va_list     ap;

    va_start(ap,First);
    vfprintf(bugfile,First,ap);
    va_end(ap);
    fflush(bugfile);
  }
#endif


/********************************************************************************/
/* sack stuff */

#define BytesPerElement (sizeof(long))
#define MaxBytesPerBlock (BytesPerElement * 64)

#define FirstBlock (TheSack->xFirstBlock)
#define NextHandleToAccess (TheSack->xNextHandleToAccess)
#define NextIndexToAccess (TheSack->xNextIndexToAccess)


CSack*      NewCSack(void)
  {
    CSack*    TheSack;

    TheSack = (CSack*)malloc(sizeof(CSack));
    FirstBlock = (SackBlock*)malloc(sizeof(SackBlock));
    FirstBlock->Next = NIL;
    FirstBlock->NumElements = 0;
    FirstBlock->NumBytes = 0;
    ResetScan(TheSack);
    return TheSack;
  }


void        PushElement(CSack* TheSack, void* Element)
  {
    SackBlock*    Temp;
    short         Scan;

    if (((*FirstBlock).NumBytes) < MaxBytesPerBlock)
      {
        Temp = (SackBlock*)malloc(((*FirstBlock).NumBytes
          + BytesPerElement) + sizeof(SackBlock));
        (*Temp).Next = (*FirstBlock).Next;
        (*Temp).NumElements = (*FirstBlock).NumElements + 1;
        (*Temp).NumBytes = (*FirstBlock).NumBytes + BytesPerElement;
        memcpy(&((*Temp).Data[BytesPerElement]),&((*FirstBlock).Data[0]),
          (*FirstBlock).NumBytes);
        memcpy(&((*Temp).Data[0]),Element,BytesPerElement);
        free(FirstBlock);
        FirstBlock = Temp;
      }
     else
      {
        Temp = (SackBlock*)malloc(sizeof(SackBlock) + BytesPerElement);
        (*Temp).Next = FirstBlock;
        (*Temp).NumElements = 1;
        (*Temp).NumBytes = BytesPerElement;
        memcpy(&((*Temp).Data[0]),Element,BytesPerElement);
        FirstBlock = Temp;
      }
  }


void        ResetScan(CSack* TheSack)
  {
    NextHandleToAccess = FirstBlock;
    NextIndexToAccess = -BytesPerElement; /* start out one before */
  }


MyBoolean   GetNext(CSack* TheSack, void* PlaceToPut)
  {
    short     Scan;

   RestartPoint:
    if (NextHandleToAccess == NIL)
      {
        return False;
      }
    NextIndexToAccess += BytesPerElement;
    if (NextIndexToAccess >= (*NextHandleToAccess).NumBytes)
      {
        NextHandleToAccess = (*NextHandleToAccess).Next;
        NextIndexToAccess = -BytesPerElement;
        goto RestartPoint;
      }
    memcpy(PlaceToPut,&((*NextHandleToAccess).Data[NextIndexToAccess]),BytesPerElement);
    return True;
  }


MyBoolean   KillElement(CSack* TheSack, void* Element)
  {
    SackBlock*      Scan;
    SackBlock*      Lag;
    short           Index;

    Scan = FirstBlock;
    Lag = NIL;
    while (Scan != NIL)
      {
        for (Index = 0; Index < (*Scan).NumBytes; Index += BytesPerElement)
          {
            if (0 == memcmp(Element,&((*Scan).Data[Index]),BytesPerElement))
              {
                SackBlock*    Temp;
                short         Count;

                Temp = (SackBlock*)malloc((*Scan).NumBytes
                  - BytesPerElement + sizeof(SackBlock));
                (*Temp).Next = (*Scan).Next;
                (*Temp).NumElements = (*Scan).NumElements - 1;
                (*Temp).NumBytes = (*Scan).NumBytes - BytesPerElement;
                memcpy(&((*Temp).Data[0]),&((*Scan).Data[0]),Index);
                memcpy(&((*Temp).Data[Index]),&((*Scan).Data[Index + BytesPerElement]),
                  (*Scan).NumBytes - BytesPerElement - Index);
                if (Lag == NIL)
                  {
                    FirstBlock = Temp;
                  }
                 else
                  {
                    (*Lag).Next = Temp;
                  }
                free(Scan);
                if (((*Temp).NumElements == 0) && (Temp != FirstBlock))
                  {
                    (*Lag).Next = (*Temp).Next;
                    free(Temp);
                  }
                return True; /* return after deletion of one */
              }
          }
        Lag = Scan;
        Scan = (*Scan).Next;
      }
    return False;
  }


long        NumElements(CSack* TheSack)
  {
    long        Accr;
    SackBlock*  Scan;

    Accr = 0;
    Scan = FirstBlock;
    while (Scan != NIL)
      {
        Accr += (*Scan).NumElements;
        Scan = (*Scan).Next;
      }
    return Accr;
  }
