/* CommonStuff.h */

/* comment out this line to turn off audit file generation */
/* #define dBUG */

#include <errno.h>
#include <sys/types.h>
#include <signal.h>
#include <fcntl.h>
#include <ttyent.h>
#include <string.h>
#include <stdarg.h>
#include <setjmp.h>
#include <sgtty.h>
#include <stdio.h>
#include <sys/file.h>
#include <sys/stat.h>


typedef short MyBoolean;

#define True (0==0)
#define False (0!=0)
#define NIL ((void*)0)

#ifndef compilingmain
extern volatile unsigned long TickTrack;
#endif

#define TickCount() (TickTrack)

#ifndef compilingcommonstuff
extern volatile MyBoolean   QuitFlag;
#endif

/* debugging macro */
#ifdef dBUG
#define APRINT(param) AuditPrint param
#define ERROR(cond,cmd) if (cond) cmd
void    AuditPrint(char* First, ...);
#else
#define APRINT(param)
#define ERROR(cond,cmd)
#endif


typedef struct
  {
    struct sgttyb     TTYB;
    struct tchars     TCHARS;
    struct ltchars    LTCHARS;
    int               LDISC;
    int               LMODE;
    long              fcntlFlags;
  } TTYInfo;


extern TTYInfo    OriginalTTY;
#ifdef dBUG
extern FILE*      bugfile;
#endif

void      GetTTYInfo(int TheFile, TTYInfo* TheInfo);
void      SetTTYInfo(int TheFile, TTYInfo* TheInfo);
void      MakeSpiffyTTY(TTYInfo* TheInfo);


extern int errno;

typedef unsigned long ulong;
typedef unsigned char uchar;

#define AbortFlag (False)
#define DoneFlag (True)

#define ABORTTHRESHHOLD (5) /* number of cancels's needed to abort transfer */
#define CANCELCHARACTER (24) /* ASCII code for cancel character */

/* amount of data in a packet.  Must not exceed maximum packet size from CMyDocument */
#define FileXferPacketDataSize (1024)

/* this is the length of time in between retries, in 60ths of a second */
#define RetryIntersticeDelay (60*3)

/* file transfer protocol messages */

#define HeaderBlockWord (0x0000)
#define HeaderBlockFileLengthPlace (1) /* 4 bytes */
#define HeaderBlockFileNamePlace (5) /* 33 chars--it's a pascal string */
#define HeaderBlockLength (38)

#define DataBlockWord (0x0001)
#define DataBlock1KIndex (1) /* 3 bytes indicating what 1k block the data is for */
#define DataBlockDataStart (4)
#define DataBlockLength (1 + 3) /* add FileXferPacketDataSize to it */

#define ResendRequestWord (0x0002)
#define ResendRequest1KIndex (1) /* 3 bytes as above */
#define ResendRequestLength (4)

#define GoodGotoNextFileWord (0x0003)
#define GoodGotoNextFileLength (1)

#define GotHeaderStartSendingWord (0x0004)
#define GotHeaderStartSendingLength (1)

#define DoneSendingResolveErrorsWord (0x0007)
#define DoneSendingResolveErrorsLength (1)


#define MaxTotalPacketSize (FileXferPacketDataSize * 2) /* plenty */

typedef struct TransRec
  {
    char*       Data;
    short       Length;
  } TransRec;


void        ErrorEncode(TransRec* Packet);
MyBoolean   ErrorDecode(char* Data, short* Length);
void        TransmitEncode(TransRec* Packet, short NumBits);
void        TransmitEncode8(TransRec* Packet);
void        TransmitEncode7(TransRec* Packet);
MyBoolean   TransmitDecode(char* Data, short* Length, short NumBits);
MyBoolean   TransmitDecode8(char* Data, short* Length);
MyBoolean   TransmitDecode7(char* Data, short* Length);


typedef struct SackBlock
  {
    struct SackBlock*   Next;
    short               NumElements;
    short               NumBytes;
    char                Data[1];
  } SackBlock;

typedef struct
  {
    SackBlock*  xFirstBlock;
    SackBlock*  xNextHandleToAccess;
    short       xNextIndexToAccess;
  } CSack;

CSack*      NewCSack(void);
void        PushElement(CSack* TheSack, void* Element);
void        ResetScan(CSack* TheSack);
MyBoolean   GetNext(CSack* TheSack, void* Element);
MyBoolean   KillElement(CSack* TheSack, void* Element);
long        NumElements(CSack* TheSack);

