/* CMyDocument.c */

#define CompilingCMyDocument

#include "control.h"
#include "crc32.h"

/* how big our return packet will be compared to the one we */
/* just received. */
#define PACKETSCALINGRATIO(size) (((size * 9) / 10) - 10)

/* won't transmit packets smaller than this */
#define ABSMINPACKETSIZE (30)

/* high level routing of data */
CTerminal*    Handlers[MaxNumSessions]; /* NIL == not in use */
BuffRec       Buffers[MaxNumSessions];
char          MakeBuffers[MaxNumSessions][AbsoluteMaxPacketDataSize];

/* low level lost packet replacement things */
TransRec      ReceivePackets[MaxPendingPackets];
TransRec      SendPackets[MaxPendingPackets];
short         ReceiveBasePacketIDNum;
short         ReceiveNumPackets;
short         SendBasePacketIDNum;
short         SendNumPackets;

/* Session 0 is special.  Its CTerminal Handler is the root console, but it */
/* does not receive data unless DataMode == SingleSession, in which case it */
/* sends raw data.  Data received from session 0 is for packet validation and */
/* commands.  The other sessions are normal */
char          PartialPacketArray[READBUFSIZE]; /* DocumentIdle stores partial packets here */
short         PartialPacketLength;
long          TimeOutFactor;
char          ChannelZeroBuffer[MaxChannelZeroCmdSize];
short         ChannelZeroBufPtr;
short         EffectiveMaxPacketSize;

short         NumDataBits;

MyBoolean     CanTransmitFlag = False;

volatile MyBoolean  QuitFlag = False;
volatile long	TickTrack = 0;


/* here is some extra stuff: the PTY buffer can only hold about 256 bytes, so */
/* we have to buffer the written data ourselves. */
long          GlobalBufStart = 0;
long          GlobalBufEnd = 0;
#define GlobalBufSize (4096)
#define GlobalBufMask (GlobalBufSize - 1)
char          GlobalBuffer[GlobalBufSize];



/* the following values are defined for portability. */
#define MAXPACKETID (16383)
#define MINPACKETID (-16383)
#define STARTINGPACKETID (-100)


/* channel zero messages: */

#define NewSessionRequestMsgWord (0x0080)
#define NewSessionRequestMsgLength (1)

#define NewSessionGrantedMsgWord (0x0081)
#define NewSessionGrantedMsgLength (2)
#define NewSessionGrantedChannelIDIndex (1)

#define NewSessionDeniedMsgWord (0x0082)
#define NewSessionDeniedMsgLength (1)

#define CloseSessionCommandMsgWord (0x0083)
#define CloseSessionCommandMsgLength (2)
#define CloseSessionCommandChannelIDIndex (1)

#define KillServerCommandMsgWord (0x0084)
#define KillServerCommandMsgLength (1)

#define WindowRowHighBit (0x01)
#define WindowColumnHighBit (0x02)

#define WindowSizeChangeRequestWord (0x0088)
#define WindowSizeChangeRequestSessionSelector (1)
#define WindowSizeChangeRequestRowSelector (2)
#define WindowSizeChangeRequestColumnSelector (3)
#define WindowSizeChangeRequestHighBits (4)
#define WindowSizeChangeRequestLength (5)

#define WindowSizeChangeCommandWord (0x0089)
#define WindowSizeChangeCommandSessionSelector (1)
#define WindowSizeChangeCommandRowSelector (2)
#define WindowSizeChangeCommandColumnSelector (3)
#define WindowSizeChangeCommandHighBits (4)
#define WindowSizeChangeCommandLength (5)

#define RequestOpenWindowInfoCmd (0x008a)
#define RequestOpenWindowInfoLength (1)

#define SendOpenWindowInfoCmd (0x008b)
#define SendOpenWindowInfoCount (1)
#define SendOpenWindowInfoCellSize (4)
#define SendOpenWindowInfoCellOffset (2)
#define SendOpenWindowInfoIndexOffset (0)
#define SendOpenWindowInfoRowOffset (1)
#define SendOpenWindowInfoColumnOffset (2)
#define SendOpenWindowInfoHighBitsOffset (3)
/* this is a special command.  There are 4*count bytes, each one specifying */
/* the index number of the open window, the number of rows in it, the next */
/* number of columns in it, and the last specifying the high bits */
/* This command should NEVER be received by the server.  In other words, */
/* the server should never send "RequestOpenWindowInfoCmd" */


/* initialize internal registers */
void      InitializeCMyDocument(void)
  {
    short     Scan;
    char*     TempPtr;

    APRINT(("+InitializeCMyDocument\n"));
    PartialPacketLength = 0;
    ChannelZeroBufPtr = 0;
    ReceiveBasePacketIDNum = STARTINGPACKETID;
    ReceiveNumPackets = 0;
    SendBasePacketIDNum = STARTINGPACKETID;
    SendNumPackets = 0;
    EffectiveMaxPacketSize = ABSMINPACKETSIZE;
    /* timeout == 1.5 times the length of time a packet ought to take */
    TimeOutFactor = (MaxPendingPackets * (60/4) * 3) / 2;
    /* delay between packets is always 15/60ths of a second */

    /* clearing handler table and data buffers */
    for (Scan = 0; Scan < MaxNumSessions; Scan += 1)
      {
        Handlers[Scan] = NIL;
        Buffers[Scan].DataStart = 0;
        Buffers[Scan].DataEnd = 0;
      }

    /* allocating packet buffers */
    for (Scan = 0; Scan < MaxPendingPackets; Scan += 1)
      {
        ReceivePackets[Scan].Data = malloc(MaxTotalPacketSize);;
        ReceivePackets[Scan].Length = -1;
        ReceivePackets[Scan].LastActionTime = TickCount();
        SendPackets[Scan].Data = malloc(MaxTotalPacketSize);;
        SendPackets[Scan].Length = -1;
        SendPackets[Scan].Acknowledged = False;
        SendPackets[Scan].LastActionTime = TickCount();
      }

    APRINT(("-InitializeCMyDocument\n"));
  }


void      ShutDownCMyDocument(void)
  {
    short         Scan;

    APRINT(("+ShutDownCMyDocument\n"));
    for (Scan = 0; Scan < MaxNumSessions; Scan += 1)
      {
        if (Handlers[Scan] != NIL)
          {
            DeletePty(Handlers[Scan]);
          }
      }
    APRINT(("-ShutDownCMyDocument\n"));
  }


/*   CTerminals use this to send data to the remote socket.  If there is no room */
/* available in the pending packet list, then it returns FALSE, otherwise, if */
/* the packet was posted successfully, then it returns TRUE.  It does not spin-wait */
/* and it's the CTerminal's problem if data wasn't successfully posted.  */
/* CTerminal should be posting data during it's DocumentIdle loop and should just exit */
/* and try again next time. */
/*   If a packet is posted successfully, then its data is encoded for error checking */
/* and transport over 7 or 8 bit lines, and dumped to the modem. */
MyBoolean   ReceiveDataFromLocalTerminal(char* Data, short Length, CTerminal* Sender)
  {
    short     SessionNumber;

    /* finding out who send the packet. */
    SessionNumber = 0;
    while ((SessionNumber < MaxNumSessions) && (Handlers[SessionNumber] != Sender))
      {
        SessionNumber += 1;
      }

    return PostPacket(Data,Length,SessionNumber);
  }


/* this routine posts data to be sent via one of the channels. */
MyBoolean   PostPacket(char* Data, short Length, short SessionNumber)
  {
    short     Scan;
    short     DataStart;
    short     DataEnd;
    char*     DataPtr;

    /* if there is enough room, queue the data into the session's buffer */
    DataStart = Buffers[SessionNumber].DataStart;
    DataEnd = Buffers[SessionNumber].DataEnd;
    DataPtr = Buffers[SessionNumber].Data;
    if ((((DataEnd - DataStart) & BuffMask) + Length) < BuffSize - 1)
      {
        /* there is enough room to put data in buffer */
        for (Scan = 0; Scan < Length; Scan += 1)
          {
            DataPtr[DataEnd] = Data[Scan];
            DataEnd = (DataEnd + 1) & BuffMask;
          }
        Buffers[SessionNumber].DataEnd = DataEnd; /* restore the values we changed. */
        return True;
      }
     else
      {
        return False;
      }
  }


void      DocumentIdle(void)
  {
    char        ReadBuffer[READBUFSIZE];
    long        NumBytesInBuffer;
    long        FileBytes;
    long        Scan;
    short       Index;
    long        WriteBytes;
    long        WriteReturn;
    char        WriteLocalBuff[64];
    long        ShadowGlobalBufStart;
    MyBoolean   GotAPacket;

    /* first, we write any data in the buffer that we can. */
   WriteSomeMore:
    WriteBytes = 0;
    ShadowGlobalBufStart = GlobalBufStart;
    while ((WriteBytes < 64) && (ShadowGlobalBufStart != GlobalBufEnd))
      {
        WriteLocalBuff[WriteBytes] = GlobalBuffer[ShadowGlobalBufStart];
        ShadowGlobalBufStart = (ShadowGlobalBufStart + 1) & GlobalBufMask;
        WriteBytes += 1;
      }
    if (WriteBytes > 0)
      {
        /* we get here if there actually is some data to write */
        WriteReturn = write(1,WriteLocalBuff,WriteBytes);
        if (WriteReturn > 0)
          {
            /* we get here if the write succeeded at all */
            if (WriteReturn == WriteBytes)
              {
                /* entire write succeeded, delete all data */
                GlobalBufStart = ShadowGlobalBufStart;
                goto WriteSomeMore; /* we'll write the whole buffer if we can */
              }
             else
              {
                /* partial write succeeded, delete whatever was sent */
                while (WriteReturn > 0)
                  {
                    GlobalBufStart = (GlobalBufStart + 1) & GlobalBufMask;
                  }
              }
          }
      } /* end of 'there was some data to be written' */

    GotAPacket = False;
   ReadChannelPart:
    /* copying partial packet from last time into buffer */
    NumBytesInBuffer = PartialPacketLength;
    memcpy(ReadBuffer,PartialPacketArray,NumBytesInBuffer);
   ScanMoreStuff:
    /* trying to fill rest of buffer with modem data. */
    /* first, get number of bytes in buffer so we can do our reads non-blockingly */
    FileBytes = READBUFSIZE - NumBytesInBuffer;
    FileBytes = read(0,&(ReadBuffer[NumBytesInBuffer]),FileBytes);
    if (FileBytes < 0)
      {
        FileBytes = 0;
      }
    NumBytesInBuffer += FileBytes; /* add in file bytes */
    Scan = 0;
    /* now we will scan for a 0x0d to find the end of a packet. */
    while ((Scan < NumBytesInBuffer) && (ReadBuffer[Scan] != 0x0d))
      {
        Scan += 1;
      }
    if (Scan == READBUFSIZE)
      {
        /* this means there is a packet that's bigger than it possibly could be */
        /* so we'll just discard it */
        NumBytesInBuffer = 0;
        goto ScanMoreStuff;
      }
    if (Scan == NumBytesInBuffer)
      {
        /* we want to save whatever data we got. */
        /* even if we read some data last time through, if we didn't read enough */
        /* that a packet break was found, we probably won't get any more */
        /* so we'll just save the data and exit. */
       CleanUpRead:
        PartialPacketLength = NumBytesInBuffer;
        memcpy(PartialPacketArray,ReadBuffer,NumBytesInBuffer);
        goto WriteChannelPart;
      }
     else
      {
        short       LengthTemp;
        short       PacketID;

        /* otherwise we've found a valid packet, so we'll try to decode it. */
        /* we don't have to allocate another buffer, since decoding data ALWAYS */
        /* makes it smaller. */
        LengthTemp = Scan;

        if (!TransmitDecode(ReadBuffer,&LengthTemp))
          {
            APRINT(("transmit decode failed\n"));
            goto ThrowAwayPacket;
          }
        if (!ErrorDecode(ReadBuffer,&LengthTemp))
          {
            APRINT(("error decode failed\n"));
            goto ThrowAwayPacket;
          }
        PacketID = (unsigned char)ReadBuffer[0]
          + (((signed short)(signed char)ReadBuffer[1]) * 256);
        Index = PMod(PacketID - ReceiveBasePacketIDNum);

        if (LengthTemp == 2)
          {
            /* it's an acknowledge packet */
            APRINT(("got ack for %d\n",(int)PacketID));
            DoAcknowledgePacket(PacketID);
            goto ThrowAwayPacket;
          }

        if ((LengthTemp == 3) && ((unsigned char)ReadBuffer[2] == 0x00ff))
          {
            /* its a request for a packet */
            Index = PMod(PacketID - SendBasePacketIDNum);
            if ((Index >= 0) && (Index < SendNumPackets)
              && (!SendPackets[Index].Acknowledged)
              && (SendPackets[Index].Length != -1))
              {
                /* we can resend packet.  Actually, that last test */
                /* (Length != -1) really shouldn't be necessary */
                SendPackets[Index].LastActionTime = TickCount() - TimeOutFactor;
                /* we make it so the packet looks like it's timed-out, so */
                /* it will get resent at the next available opportunity */
              }
            goto ThrowAwayPacket;
          }

        /* since we have no idea how much data is in our buffer waiting to */
        /* be transmitted, we use the other side which does know how much data */
        /* is in it's buffer.  We assume bandwidth in both directions is equal */
        /* (maybe not a good idea) and send packets matching ones we receive. */
        /* (one response per receipt, and same raw data size.) */
        /* For this to work at all well, this routine must be invoked more */
        /* frequently than complete packets arrive. */
        CanTransmitFlag = True;
        EffectiveMaxPacketSize = PACKETSCALINGRATIO(LengthTemp);
        if (NumDataBits == 7)
          {
            /* empirically, 7-bit encoding seems to need a little more */
            /* breathing room, so we'll knock it down some more. */
            EffectiveMaxPacketSize = (7 * EffectiveMaxPacketSize) / 8;
          }
        if (EffectiveMaxPacketSize < ABSMINPACKETSIZE)
          {
            EffectiveMaxPacketSize = ABSMINPACKETSIZE;
          }
        if (EffectiveMaxPacketSize > AbsoluteMaxPacketDataSize)
          {
            EffectiveMaxPacketSize = AbsoluteMaxPacketDataSize;
          }

        /* putting packet into received packet buffer */
        if (Index < 0)
          {
            APRINT(("redundant %d\n",(int)PacketID));
            SendAcknowledgePacket(PacketID);
            goto ThrowAwayPacket;
          }
        if (Index >= MaxPendingPackets)
          {
            /* Received packet with ID above range we can handle. */
            APRINT(("received early packet %d\n",(int)PacketID));
            goto ThrowAwayPacket;
          }
        if (ReceivePackets[Index].Length != -1)
          {
            /* received duplicate packet */
            APRINT(("received duplicate %d\n",(int)PacketID));
            SendAcknowledgePacket(PacketID);
            goto ThrowAwayPacket;
          }
        if (!CheckPacketFormat(ReadBuffer,LengthTemp))
          {
            APRINT(("bad packet format %d\n",(int)PacketID));
            goto ThrowAwayPacket;
          }
        if (ReceiveNumPackets < Index + 1)
          {
            short     Scan;

            APRINT(("rcvd %d\n",(int)PacketID));
            /* this means we are receiving a new packet, not a packet we */
            /* should have received before.  We check for any packets */
            /* between the new one and the previous one which weren't */
            /* received (they should be in order.  And this won't get invoked */
            /* for resent packets) and request any missing packets, which are */
            /* assumed to have been discarded due to line noise */
            for (Scan = ReceiveNumPackets; Scan < Index; Scan += 1)
              {
                SendPacketRequest(PMod(Scan + ReceiveBasePacketIDNum));
              }
            ReceiveNumPackets = Index + 1;
          }
        memcpy(ReceivePackets[Index].Data,ReadBuffer,LengthTemp);
        ReceivePackets[Index].Length = LengthTemp;
        SendAcknowledgePacket(PacketID);
        goto ThrowAwayPacket;

        /* no matter what, we want to flush that packet from the read buffer */
        /* and try to decode the next one. */
       ThrowAwayPacket:
        /* NumBytesInBuffer - Scan - 1 is the number of bytes in the buffer after */
        /* the packet [scan] and it's terminator char (1) have been subtracted. */
        for (Index = 0; Index < (NumBytesInBuffer - Scan - 1); Index += 1)
          {
            ReadBuffer[Index] = ReadBuffer[Index + Scan + 1];
          }
        NumBytesInBuffer -= Scan + 1;  /* packet & terminator have been removed. */
        goto ScanMoreStuff;
      }

   WriteChannelPart:
    /* first, we see how many received packets we can process */
    while (ReceivePackets[0].Length != -1)
      {
        char*     PtrSave;

        /* use the packet's data (send to terminals) */
        UnmakePacket(ReceivePackets[0].Data,ReceivePackets[0].Length);
        GotAPacket = True;
        /* the loop will destroy the data block in the lowest record, */
        /* so we preserve it and put it back on the top one. */
        PtrSave = ReceivePackets[0].Data;
        for (Index = 0; Index < ReceiveNumPackets - 1; Index += 1)
          {
            /* shift them all down. */
            ReceivePackets[Index] = ReceivePackets[Index + 1];
          }
        ReceiveNumPackets -= 1;
        ReceiveBasePacketIDNum = PMod(ReceiveBasePacketIDNum + 1);
        ReceivePackets[ReceiveNumPackets].Data = PtrSave;
        ReceivePackets[ReceiveNumPackets].Length = -1;
        ReceivePackets[ReceiveNumPackets].LastActionTime = TickCount();
      }

    /* request any packets we haven't received for a while */
    for (Index = 0; Index < MaxPendingPackets; Index += 1)
      {
        if ((TickCount () - ReceivePackets[Index].LastActionTime)
          > (TimeOutFactor * 2))
          {
            SendPacketRequest(PMod(Index + ReceiveBasePacketIDNum));
            ReceivePackets[Index].LastActionTime = TickCount();
          }
      }

    /* if there's room in the buffer, then we'll send another packet */
   LoopPoint:
    if ((!GotAPacket) && CanWeSendPacket())
      {
        /* what packet will we send?  If there is one that has not been */
        /* acknowledged and the timeout period has expired, we'll resend */
        /* that one, otherwise we'll make a new one and send that */
        for (Index = 0; Index < SendNumPackets; Index += 1)
          {
            if ((!SendPackets[Index].Acknowledged)
              && ((signed long)TickCount() - SendPackets[Index].LastActionTime
                >= TimeOutFactor))
              {
                WriteBlock(SendPackets[Index].Data,SendPackets[Index].Length);
                CanTransmitFlag = False;
                APRINT(("resent %d\n",(int)PMod(Index + SendBasePacketIDNum)));
                SendPackets[Index].LastActionTime = TickCount();
                goto LoopPoint;
              }
          }
        if (SendNumPackets < MaxPendingPackets)
          {
            /* if there is still space in the table to make a new packet, */
            /* then we will. */
            MakePacket(PMod(SendBasePacketIDNum + SendNumPackets),
              &(SendPackets[SendNumPackets]));
            WriteBlock(SendPackets[SendNumPackets].Data,
              SendPackets[SendNumPackets].Length);
            CanTransmitFlag = False;
            APRINT(("sent %d\n",(int)PMod(SendNumPackets + SendBasePacketIDNum)));
            SendPackets[SendNumPackets].LastActionTime = TickCount();
            SendPackets[SendNumPackets].Acknowledged = False;
            SendNumPackets += 1;
            goto LoopPoint;
          }
      }
  }


void      DoAcknowledgePacket(short PacketID)
  {
    short     Index;
    char*     BufferSave;

    Index = PMod(PacketID - SendBasePacketIDNum);
    if (Index < 0)
      {
        return; /* means we've already passed that packet up */
      }
    if (Index > SendNumPackets)
      {
        return; /* means we haven't send that packet yet */
      }
    SendPackets[Index].Acknowledged = True;
    while (SendPackets[0].Acknowledged)
      {
        /* if the first packet has been acknowledged, we can get rid of it and */
        /* shift all of the other packets down. */
        BufferSave = SendPackets[0].Data; /* save AllocPtr'd buffer */
        for (Index = 0; Index < SendNumPackets - 1; Index += 1)
          {
            SendPackets[Index] = SendPackets[Index + 1];
          }
        SendNumPackets -= 1;
        SendBasePacketIDNum = PMod(SendBasePacketIDNum + 1);
        SendPackets[SendNumPackets].Acknowledged = False;
        SendPackets[SendNumPackets].Data = BufferSave;
        SendPackets[SendNumPackets].Length = -1;
      }
  }


void      ReceiveChannelZeroCommand(char* Data, short Length)
  {
   PlopPoint:
    while (Length > 0)
      {
        ChannelZeroBuffer[ChannelZeroBufPtr] = Data[0];
        Data += 1;
        ChannelZeroBufPtr += 1;
        Length -= 1;
        if ((ChannelZeroBufPtr == 1) && ((unsigned char)ChannelZeroBuffer[0] < 0x0080))
          {
            ChannelZeroBufPtr = 0;
            goto PlopPoint;
          }
        switch ((unsigned char)ChannelZeroBuffer[0])
          {
            case NewSessionRequestMsgWord:
              if (ChannelZeroBufPtr == NewSessionRequestMsgLength)
                {
                  DoNewSessionRequest();
                  ChannelZeroBufPtr = 0;
                }
              break;
            case NewSessionGrantedMsgWord:
              if (ChannelZeroBufPtr == NewSessionGrantedMsgLength)
                {
                  DoNewSessionGranted(ChannelZeroBuffer[NewSessionGrantedChannelIDIndex]);
                  ChannelZeroBufPtr = 0;
                }
              break;
            case NewSessionDeniedMsgWord:
              /* server shouldn't receive this since it should never request a session */
              if (ChannelZeroBufPtr == NewSessionDeniedMsgLength)
                {
                  ChannelZeroBufPtr = 0;
                }
              break;
            case CloseSessionCommandMsgWord:
              if (ChannelZeroBufPtr == CloseSessionCommandMsgLength)
                {
                  DoCloseSessionCommand(ChannelZeroBuffer[CloseSessionCommandChannelIDIndex]);
                  ChannelZeroBufPtr = 0;
                }
              break;
            case KillServerCommandMsgWord:
              if (ChannelZeroBufPtr == KillServerCommandMsgLength)
                {
                  ChannelZeroBufPtr = 0;
                  QuitFlag = True;
                }
              break;
            case WindowSizeChangeCommandWord:
              /* this is in response to a windo size change demand. */
              /* we can't handle it */
              if (ChannelZeroBufPtr == WindowSizeChangeCommandLength)
                {
                  ChannelZeroBufPtr = 0;
                }
              break;
            case WindowSizeChangeRequestWord:
              /* this is when the other side requests a change. We echo a change back */
              if (ChannelZeroBufPtr == WindowSizeChangeRequestLength)
                {
                  char      EchoBuf[WindowSizeChangeCommandLength];

#ifdef TIOCSWINSZ
                  if (Handlers[ChannelZeroBuffer[WindowSizeChangeRequestSessionSelector]]
                    != NIL)
                      {
                        short   XTemp,YTemp;

                        XTemp = ChannelZeroBuffer[WindowSizeChangeRequestColumnSelector];
                        if ((ChannelZeroBuffer[WindowSizeChangeRequestHighBits]
                          & WindowColumnHighBit) != 0)
                          {
                            XTemp |= 0x80;
                          }
                        YTemp = ChannelZeroBuffer[WindowSizeChangeRequestRowSelector];
                        if ((ChannelZeroBuffer[WindowSizeChangeRequestHighBits]
                          & WindowRowHighBit) != 0)
                          {
                            YTemp |= 0x80;
                          }
                        WindowSizeChange(Handlers[ChannelZeroBuffer[WindowSizeChangeRequestSessionSelector]],
                          XTemp,YTemp);
                        EchoBuf[0] = WindowSizeChangeCommandWord;
                        EchoBuf[WindowSizeChangeRequestSessionSelector]
                          = ChannelZeroBuffer[WindowSizeChangeRequestSessionSelector];
                        EchoBuf[WindowSizeChangeCommandRowSelector]
                          = ChannelZeroBuffer[WindowSizeChangeRequestRowSelector];
                        EchoBuf[WindowSizeChangeCommandColumnSelector]
                          = ChannelZeroBuffer[WindowSizeChangeRequestColumnSelector];
                        EchoBuf[WindowSizeChangeCommandHighBits]
                          = ChannelZeroBuffer[WindowSizeChangeRequestHighBits];
                        SendChannelZeroCommand(EchoBuf,WindowSizeChangeCommandLength);
                      }
#endif
                    ChannelZeroBufPtr = 0;
                  }
              break;
            case RequestOpenWindowInfoCmd:
              if (ChannelZeroBufPtr == RequestOpenWindowInfoLength)
                {
                  char        Temp;
                  short       Count;
                  short       Scan;

                  /* we now have to tell them how many windows and their sizes. */
                  Temp = SendOpenWindowInfoCmd;
                  SendChannelZeroCommand(&Temp,1);
                  Count = 0;
                  for (Scan = 1; Scan < MaxNumSessions; Scan += 1)
                    {
                      if (Handlers[Scan] != NIL)
                        {
                          Count += 1;
                        }
                    }
                  Temp = Count;
                  SendChannelZeroCommand(&Temp,1);
                  for (Scan = 1; Scan < MaxNumSessions; Scan += 1)
                    {
                      if (Handlers[Scan] != NIL)
                        {
                          Temp = Scan;
                          SendChannelZeroCommand(&Temp,1);
#ifdef TIOCSWINSZ
                          {
                            struct winsize    MyWindowSize;

                            ioctl(Handlers[Scan]->MasterDesc,
                              TIOCGWINSZ,&MyWindowSize);
                            Temp = MyWindowSize.ws_row & 0x7f;
                            SendChannelZeroCommand(&Temp,1);
                            Temp = MyWindowSize.ws_col & 0x7f;
                            SendChannelZeroCommand(&Temp,1);
                            Temp = 0;
                            if ((MyWindowSize.ws_row & 0x80) != 0)
                              {
                                Temp |= WindowRowHighBit;
                              }
                            if ((MyWindowSize.ws_col & 0x80) != 0)
                              {
                                Temp |= WindowColumnHighBit;
                              }
                            SendChannelZeroCommand(&Temp,1);
                          }
#else
                          Temp = 24;
                          SendChannelZeroCommand(&Temp,1);
                          Temp = 80;
                          SendChannelZeroCommand(&Temp,1);
                          Temp = 0;
                          SendChannelZeroCommand(&Temp,1);
#endif
                        }
                    }
                  ChannelZeroBufPtr = 0;
                }
              break;
            default:
              ChannelZeroBufPtr = 0;
              APRINT(("received unknown command\n"));
              break;
          }
        if (ChannelZeroBufPtr == MaxChannelZeroCmdSize)
          {
            ChannelZeroBufPtr = 0;
            APRINT(("received unknown command\n"));
          }
      }
  }


/* 0x80 NewSessionRequest */
/*  See if there is enough room in the session table.  If there is, create a new */
/*  terminal and return a NewSessionGranted message indicating what the ID number is */
void      DoNewSessionRequest(void)
  {
    short   SessionScan;
    char    Deny[] = {NewSessionDeniedMsgWord};

    APRINT(("+DoNewSessionRequest\n"));
    SessionScan = 1;
    while ((SessionScan < MaxNumSessions) && (Handlers[SessionScan] != NIL))
      {
        SessionScan += 1;
      }
    if (SessionScan >= MaxNumSessions)
      {
        /* there wasn't enough room to create a new session, so we have to send */
        /* a NewSessionDenied message */
       DenyPoint:
        SendChannelZeroCommand(Deny,NewSessionDeniedMsgLength);
        APRINT(("New session request: denied\n"));
      }
     else
      {
        char        Grant[] = {NewSessionGrantedMsgWord, 0x00};
        CTerminal*  Thing;

        /* we make the new session and send a NewSessionGranted message */
        Thing = NewPty();
        if (Thing == NIL)
          {
            goto DenyPoint; /* no more PTYs available in the system */
          }
        Handlers[SessionScan] = Thing;
        Buffers[SessionScan].DataStart = 0;
        Buffers[SessionScan].DataEnd = 0;
        Grant[NewSessionGrantedChannelIDIndex] = SessionScan;
        SendChannelZeroCommand(Grant,NewSessionGrantedMsgLength);
        APRINT(("New session request: granted %d\n",(int)SessionScan));
      }
    APRINT(("-DoNewSessionRequest\n"));
  }


/* 0x81 NewSessionGranted */
/*  Sent in response to a NewSessionRequest message.  It contains the ID number */
/*  of the new session.  If we receive it, then we create a new terminal with that */
/*  ID number. (we don't return anything.) */
void      DoNewSessionGranted(short SessionID)
  {
    CTerminal*  Thing;

    APRINT(("+DoNewSessionGranted\n"));
    /* if there is an available pty, we obey by opening it, otherwise we */
    /* force the remote system's window to close immediately. */
    Thing = NewPty();
    if (Thing != NIL)
      {
        if ((SessionID < 0) || (SessionID >= MaxNumSessions) || (Handlers[SessionID] != NIL))
          {
            DeletePty(Thing);
            return;
          }
        Handlers[SessionID] = Thing;
        Buffers[SessionID].DataStart = 0;
        Buffers[SessionID].DataEnd = 0;
        APRINT(("New session granted: PTY installed\n"));
      }
     else
      {
        char    Close[CloseSessionCommandMsgLength] = {CloseSessionCommandMsgWord,0};

        Close[CloseSessionCommandChannelIDIndex] = SessionID;
        SendChannelZeroCommand(Close,CloseSessionCommandMsgLength);
        APRINT(("New session granted: PTY failed, so we deny\n"));
      }
    APRINT(("-DoNewSessionGranted\n"));
  }


/* 0x83 CloseSessionCommand */
/*  If the user closes a session, or a process dies, this message gets sent to the */
/*  other system so that it knows what happened and that it should destroy it's */
/*  corresponding structures.  When we receive it, we delete the CTerminal in charge */
/*  of the specified session. */
void      DoCloseSessionCommand(short SessionID)
  {
    short     PacketScan;

    APRINT(("+DoCloseSessionCommand\n"));
    if ((SessionID < 0) || (SessionID >= MaxNumSessions) || (Handlers[SessionID] == NIL))
      {
        APRINT(("Close session command: session was nonexistent--ignored\n"));
      }
     else
      {
        if (Handlers[SessionID] != NIL)
          {
            DeletePty(Handlers[SessionID]);
            Handlers[SessionID] = NIL;
            APRINT(("Close session command successful\n"));
          }
      }
    APRINT(("-DoCloseSessionCommand\n"));
  }


/* send a channel zero command */
void      SendChannelZeroCommand(char* Data, short Length)
  {
    /* we pretend we are the root session terminal and dump the data */
    /* to channel 0's buffer. */
    APRINT(("+SendChannelZeroCommand\n"));
    PostPacket(Data,Length,0/*session number*/);
    APRINT(("-SendChannelZeroCommand\n"));
  }


void    ErrorEncode(TransRec* Packet)
  {
    ulong     CRC;
    short     NumBytes;

    NumBytes = Packet->Length;
    CRC = UpdateCRC(Packet->Data,NumBytes);
    Packet->Data[NumBytes + 0] = CRC >> 24;
    Packet->Data[NumBytes + 1] = CRC >> 16;
    Packet->Data[NumBytes + 2] = CRC >> 8;
    Packet->Data[NumBytes + 3] = CRC;
    Packet->Length += 4;
  }


/* Returns TRUE if data is correct or FALSE if it is bad. */
MyBoolean	ErrorDecode(char* Data, short* Length)
  {
    ulong     CRC;
    short     NumDataBytes;

    (*Length) -= 4;
    NumDataBytes = (*Length);
    CRC = UpdateCRC(Data,NumDataBytes);
    return ((NumDataBytes >= 0) && (0 == (0x000000ff & (
      (Data[NumDataBytes + 0] ^ (CRC >> 24))
      | (Data[NumDataBytes + 1] ^ (CRC >> 16))
      | (Data[NumDataBytes + 2] ^ (CRC >> 8))
      | (Data[NumDataBytes + 3] ^ CRC)))));
  }


/* Make packet takes buffered data received from local sessions and makes it */
/* into a packet. */
/* It also encodes the packet for transmission (ErrorEncode and TransmitEncode) */
/* first 2 bytes = PacketID */
/* third byte = number of segments */
/* rest of data = segments (same number as third byte) */
/* segment structure:  1 byte = destination session, 1 byte = length of data, */
/* then however many data bytes there are. */
/* if there is no data to be send, then we send nothing (0 bytes) to channel 1 */
void      MakePacket(short PacketID, TransRec* Packet)
  {
    short     MaxBytes;
    short     BytesSoFar;
    short     OldBytesSoFar;
    char*     Buffer[MaxNumSessions]; /* array of our local buffers */
    char*     BufferBase[MaxNumSessions]; /* array of global data buffers */
    short     BufPtr[MaxNumSessions];
    short     DataStart[MaxNumSessions];
    short     DataEnd[MaxNumSessions];
    short     SessionNumber[MaxNumSessions];
    short     Scan;
    short     NumSegments;

    /* calculating number of segments */
    DataStart[0] = Buffers[0].DataStart;
    DataEnd[0] = Buffers[0].DataEnd;
    SessionNumber[0] = 0;
    Buffer[0] = MakeBuffers[0];
    BufferBase[0] = Buffers[0].Data;
    BufPtr[0] = 0;
    NumSegments = 1;
    for (Scan = 1; Scan < MaxNumSessions; Scan += 1)
      {
        short     DataStartTemp;
        short     DataEndTemp;

        DataStartTemp = Buffers[Scan].DataStart;
        DataEndTemp = Buffers[Scan].DataEnd;
        if (DataStartTemp != DataEndTemp)
          {
            /* if there's data or it's channel zero, then we buffer it */
            DataStart[NumSegments] = DataStartTemp;
            DataEnd[NumSegments] = DataEndTemp;
            SessionNumber[NumSegments] = Scan;
            Buffer[NumSegments] = MakeBuffers[Scan];
            BufferBase[NumSegments] = Buffers[Scan].Data;
            BufPtr[NumSegments] = 0;
            NumSegments += 1;
          }
      }

    /* filling up data */
    /* for max bytes, we doc 2 bytes for PacketID, 1 byte for the number of segments, */
    /* and 2 bytes for each segment header we'll have. */
    MaxBytes = EffectiveMaxPacketSize - 2 - 1 - (2 * NumSegments);
    BytesSoFar = 0;
    do
      {
        OldBytesSoFar = BytesSoFar;
        for (Scan = 0; Scan < NumSegments; Scan += 1)
          {
            /* if a session has any bytes waiting to be sent, then pluck one out. */
            /* also, since length is encoded as a byte + nybble, we can't exceed 4095 */
            if ((DataStart[Scan] != DataEnd[Scan]) && (BufPtr[Scan] < (255 + 256*15)))
              {
                Buffer[Scan][BufPtr[Scan]] = BufferBase[Scan][DataStart[Scan]];
                BufPtr[Scan] += 1;
                DataStart[Scan] = (DataStart[Scan] + 1) & BuffMask;
                BytesSoFar += 1;
              }
            if (BytesSoFar == MaxBytes)
              {
                goto EscapePoint;
              }
          }
      } while (OldBytesSoFar != BytesSoFar);
    while ((BytesSoFar < MaxBytes) && (BufPtr[0] < (255 + 256*15)))
      {
        static long       Seed;

        /* pad packets to full size with relatively uncompressible data */
        Seed = Seed * 1103515245 + 12345;
        Buffer[0][BufPtr[0]] = Seed & 0x007f;
        BufPtr[0] += 1;
        BytesSoFar += 1;
      }

   EscapePoint:
    /* restore data we changed. */
    for (Scan = 0; Scan < NumSegments; Scan += 1)
      {
        Buffers[SessionNumber[Scan]].DataStart = DataStart[Scan];
      }

    if (BytesSoFar == 0)
      {
        /* nobody wanted to send anything, so make a generic packet */
        Packet->Data[2] = 1; /* number of segments */
        Packet->Data[3] = 1; /* channel number (any would do.) */
        Packet->Data[4] = 0; /* number of bytes */
        Packet->Length = 5; /* 5 total bytes including packet ID */
      }
     else
      {
        short     Index;
        short     Count;

        Packet->Length = 2 + 1 + (2 * NumSegments) + BytesSoFar;
        Packet->Data[2] = NumSegments;
        Index = 3;
        for (Scan = 0; Scan < NumSegments; Scan += 1)
          {
            /* session Scan wants to send stuff, so let it. */
            Packet->Data[Index] = SessionNumber[Scan]
              | ((0x0f00 & BufPtr[Scan]) >> 4); /* channel number + msb of length */
            Index += 1;
            Packet->Data[Index] = 0x00ff & BufPtr[Scan];  /* lsb of length */
            Index += 1;
            Count = 0;
            while (Count < BufPtr[Scan])
              {
                Packet->Data[Index] = Buffer[Scan][Count];
                Index += 1;
                Count += 1;
              }
          }
      }
    /* now, we just need to put the packet ID in */
    Packet->Data[0] = PacketID & 0x00ff;
    Packet->Data[1] = (PacketID >> 8) & 0x00ff;
    ErrorEncode(Packet);
    TransmitEncode(Packet);
  }


/* this routine takes the packet apart and dispatches the data it contains */
/* to the target handlers. */
void        UnmakePacket(char* Data, short Length)
  {
    short     Scan;
    short     Count;
    short     ByteCount;
    short     UsedByteCounter;
    short     Destination;

    UsedByteCounter = 0;
    Scan = 3;  /* start of first segment */
    for (Count = 0; Count < (unsigned char)Data[2]; Count += 1) /* scan all segments */
      {
        ByteCount = (unsigned char)Data[Scan + 1]
          + (((short)Data[Scan] & 0x00f0) << 4);
        Destination = (Data[Scan] & 0x000f);
        if (Destination == 0)
          {
            /* if it's a channel 0 command, then send it directly to channel 0 */
            ReceiveChannelZeroCommand(&(Data[Scan + 2]),ByteCount);
          }
         else
          {
            if ((Destination < MaxNumSessions) && (Handlers[Destination] != NIL)
              && (ByteCount != 0))
              {
                /* handler exists, so we can send data to it. */
                DispatchPty(Handlers[Destination],&(Data[Scan + 2]),ByteCount);
              }
          }
        Scan += ByteCount + 2;
      }
  }


/* check to see that the packet's length actually matches the data stored */
/* inside it */
MyBoolean   CheckPacketFormat(char* Data, short Length)
  {
    short     Scan;
    short     Count;
    short     ByteCount;

    Scan = 3;  /* start of first segment */
    for (Count = 0; Count < (unsigned char)Data[2]; Count += 1) /* scan all segments */
      {
        if (Scan + 1 >= Length)
          {
            return False; /* past end of array */
          }
        ByteCount = (unsigned char)Data[Scan + 1]
          + (((short)Data[Scan] << 4) & 0x0f00);
        Scan += ByteCount + 2;
      }
    return (Scan == Length); /* return status of correctness */
  }


/* when the user closes a CTerminal, it's last action is to call this routine */
/* which takes care of telling the remote system to kill it's corresponding */
/* structures. */
void      TerminalWantsToDieCloseSession(CTerminal* OneThatDied)
  {
    char    TmpCmd[CloseSessionCommandMsgLength] = {CloseSessionCommandMsgWord, 0x00};
    short   Scan;

    APRINT(("+TerminalWantsToDieCloseSession\n"));
    Scan = 0;
    while ((Scan < MaxNumSessions) && (Handlers[Scan] != OneThatDied))
      {
        Scan += 1;
      }
    if (Scan != MaxNumSessions)
      {
        DoCloseSessionCommand(Scan);
        TmpCmd[CloseSessionCommandChannelIDIndex] = Scan;
        SendChannelZeroCommand(TmpCmd,CloseSessionCommandMsgLength);
      }
    APRINT(("-TerminalWantsToDieCloseSession\n"));
  }


void      TransmitEncode(TransRec* Packet)
  {
    switch (NumDataBits)
      {
        case 8:
          TransmitEncode8(Packet);
          break;
        case 7:
          TransmitEncode7(Packet);
          break;
      }
  }


void      TransmitEncode8(TransRec* Packet)
  {
    uchar     Mask;
    short     Index;
    short     DataScan;
    short     NumResolutionBytes;

    /* setting up encoding routine. */
    Index = MaxTotalPacketSize - 1; /* start putting bit array at top of block */
    NumResolutionBytes = 1;
    Mask = 0x40; /* we don't use the 8th bit */
    Packet->Data[Index] = 0;  /* zero the first entry */

    /* stepping through all data to generate masks & data */
    DataScan = 0;
    while (DataScan < Packet->Length)
      {
        uchar     Temp;

        Temp = Packet->Data[DataScan]; /* get a byte */
        if ((Temp < 32) || (Temp >= 224))
          {
            /* there is a possibility of ambiguity, so encode a bit */
            if (Temp >= 224)
              {
                /* if it is a high value, encode a 1, leave it at 0 */
                Packet->Data[Index] |= Mask;
                Temp -= 224;  /* take modulo 224 of temp */
              }
            Mask = Mask >> 1;
            if (Mask == 0)
              {
                /* step mask to next byte */
                Mask = 0x40;
                Index -= 1; /* remember, we descend */
                NumResolutionBytes += 1;
                Packet->Data[Index] = 0; /* zero the next entry */
              }
          }
        Packet->Data[DataScan++] = Temp + 32;  /* store byte into output stream. */
      }
    /* At exit, if Mask = 0x40, then Index points to the byte after the last, otherwise */
    /* it points to the last byte of data. */
    if (Mask == 0x40)
      {
        Index += 1;  /* don't include last byte if it wasn't actually used. */
        NumResolutionBytes -= 1;
      }
    /* now we have to copy the list of bits down on top of the data */
    for (DataScan = 0; DataScan < NumResolutionBytes; DataScan += 1)
      {
        Packet->Data[Packet->Length] = Packet->Data[Index++] + 32;
        Packet->Length += 1;
      }
    Packet->Data[Packet->Length] = 0x0d;  /* packet terminator character */
    Packet->Length += 1;
  }


/* TransmitEncode7 does something akin to what TransmitEncode8 does, */
/* but first it streches the 8-bit bytes into 7-bit bytes. */
void      TransmitEncode7(TransRec* Packet)
  {
    uchar     Mask;
    short     Index;
    short     DataScan;
    short     NumResolutionBytes;
    short     RawIndex;
    short     ExtraBits;
    char      ShadowChar;

    /* streching... */
    /* 0 bytes == 0 bits == 0.00 bytes == 0 bytes */
    /* 1 byte == 8 bits == 1.14 bytes == 2 bytes */
    /* 2 bytes == 16 bits == 2.29 bytes == 3 bytes */
    /* 3 bytes == 24 bits == 3.43 bytes == 4 bytes */
    /* 4 bytes == 32 bits == 4.57 bytes == 5 bytes */
    /* 5 bytes == 40 bits == 5.71 bytes == 6 bytes */
    /* 6 bytes == 48 bits == 6.86 bytes == 7 bytes */
    /* 7 bytes == 56 bits == 8.00 bytes == 8 bytes */
    /* 8 bytes == 64 bits == 9.14 bytes == 10 bytes */
    /* The pattern:  For each block of 7 bytes, add another byte. */
    /* if there are any left over, add yet another byte */

    /* first, copy all data to top of packet buffer */
    RawIndex = MaxTotalPacketSize - Packet->Length;
    for (Index = Packet->Length - 1; Index >= 0; Index -= 1)
      {
        Packet->Data[RawIndex + Index] = Packet->Data[Index];
      }

    /* now, we  s t r e c h */
    DataScan = 0;
    ExtraBits = 0;
    Mask = 0;
    for (Index = RawIndex; Index < MaxTotalPacketSize; Index += 1)
      {
        /* move low 7 bits over unchanged, mask out high bit */
        ShadowChar = Packet->Data[Index];
        Packet->Data[DataScan] = ShadowChar & 0x007f;
        Mask = Mask << 1; /* open a hole on the bottom of mask */
        /* notice how this is shifted left.  That means the oldest bit */
        /* will be at the top.  Therefore, the first byte's bit is at */
        /* the top and the last byte's bit is at the bottom */
        if ((ShadowChar & 0x0080) != 0)
          {
            /* if high bit is set, then set Mask bit, else leave it clear */
            Mask |= 1;
          }
        ExtraBits += 1;
        DataScan += 1;
        if (ExtraBits == 6)
          {
            /* if we have a whole byte worth of extra-bits, then we store them too */
            Packet->Data[DataScan] = Mask & 0x003f; /* 6 significant bits */
            DataScan += 1;
            ExtraBits = 0;
            Mask = 0;
          }
      }
    if (ExtraBits != 0)
      {
        /* if there were some bits left over at the end, then we have to */
        /* store them too. */
        Packet->Data[DataScan] = Mask;
        DataScan += 1;
      }
    Packet->Length = DataScan; /* this is now the new effective length */

    /* setting up encoding routine. */
    Index = MaxTotalPacketSize - 1; /* start putting bit array at top of block */
    NumResolutionBytes = 1;
    Mask = 0x20; /* we don't use the 8th or 7th bits */
    Packet->Data[Index] = 0;  /* zero the first entry */

    /* stepping through all data to generate masks & data */
    DataScan = 0;
    while (DataScan < Packet->Length)
      {
        uchar     Temp;

        Temp = Packet->Data[DataScan]; /* get a byte */
        if ((Temp < 32) || (Temp >= 96))
          {
            /* there is a possibility of ambiguity, so encode a bit */
            if (Temp >= 96)
              {
                /* if it is a high value, encode a 1, leave it at 0 */
                Packet->Data[Index] |= Mask;
                Temp -= 96;  /* take modulo 96 of temp */
              }
            Mask = Mask >> 1;
            if (Mask == 0)
              {
                /* step mask to next byte */
                Mask = 0x20;
                Index -= 1; /* remember, we descend */
                NumResolutionBytes += 1;
                Packet->Data[Index] = 0; /* zero the next entry */
              }
          }
        Packet->Data[DataScan++] = Temp + 32;  /* store byte into output stream. */
      }
    /* At exit, if Mask = 0x20, then Index points to the byte after the last, otherwise */
    /* it points to the last byte of data. */
    if (Mask == 0x20)
      {
        Index += 1;  /* don't include last byte if it wasn't actually used. */
        NumResolutionBytes -= 1;
      }
    /* now we have to copy the list of bits down on top of the data */
    for (DataScan = 0; DataScan < NumResolutionBytes; DataScan += 1)
      {
        Packet->Data[Packet->Length] = Packet->Data[Index++] + 32;
        Packet->Length += 1;
      }
    Packet->Data[Packet->Length] = 0x0d;  /* packet terminator character */
    Packet->Length += 1;
  }


/* this decodes data that has been received from the channel using the reverse of */
/* the method mentioned above.  The original handle is released.  There should NOT */
/* be a carriage return on the end at this point. */
MyBoolean	TransmitDecode(char* Data, short* Length)
  {
    switch (NumDataBits)
      {
        case 8:
          return TransmitDecode8(Data,Length);
        case 7:
          return TransmitDecode7(Data,Length);
        default:
          APRINT(("undefined num data bits (%d)\n",(int)NumDataBits));
          return False;
      }
  }


MyBoolean	TransmitDecode8(char* Data, short* Length)
  {
    short     DataIndex;
    short     BitCheckIndex;
    char      Mask;

    /* setting up the routine. */
    Mask = 0x40;
    DataIndex = 0; /* data bytes start at beginning and go forwards. */
    BitCheckIndex = (*Length) - 1; /* ambiguity resolution bytes start at end & go backwards. */

    /* scanning data and resolving it. */
    /* we go until DataIndex >= BitCheckIndex.  If all goes well, they should */
    /* end up being equal.  In the rare chance they arn't, then something is */
    /* missing in the data block, so we return False indicating an error. */
    while ((DataIndex < BitCheckIndex)
      || ((DataIndex == BitCheckIndex) && (Mask == 0x40)))
      {
        uchar     Temp;

        Temp = Data[DataIndex] - 32;
        if (Temp >= 224)
          {
            return False;  /* shouldn't ever happen! */
          }
        if (Temp < 32)
          {
            if (((Data[BitCheckIndex] - 32) & Mask) != 0)
              {
                Temp += 224; /* add in if bit was set, else leave low */
              }
            Mask = Mask >> 1;
            if (Mask == 0)
              {
                Mask = 0x40;
                BitCheckIndex -= 1;
              }
          }
        Data[DataIndex++] = Temp;
      }
    /* at exit:  EncodedData[DataIndex] should be FIRST resolution byte. */
    /* It's also the total number of data bytes. */
    /* and EncodedData[BitCheckIndex] should be FIRST resolution byte, OR */
    /* if Mask == 0x40, then it wrapped around falsly, and should be */
    /* LAST data byte. */
    if (Mask == 0x40)
      {
        BitCheckIndex += 1; /* now we fix the wrap-around. */
      }

    (*Length) = DataIndex; /* remote resolution bits from end */
    return (BitCheckIndex == DataIndex);  /* return error status. */
  }


/* this does the same sort of thing as TransmitDecode8, but unstreches */
/* the data after */
MyBoolean	TransmitDecode7(char* Data, short* Length)
  {
    short     DataIndex;
    short     BitCheckIndex;
    char      Mask;
    short     FinalIndex;
    short     Count;

    /* setting up the routine. */
    Mask = 0x20;
    DataIndex = 0; /* data bytes start at beginning and go forwards. */
    BitCheckIndex = (*Length) - 1; /* ambiguity resolution bytes start at end & go backwards. */

    /* scanning data and resolving it. */
    /* we go until DataIndex >= BitCheckIndex.  If all goes well, they should */
    /* end up being equal.  In the rare chance they arn't, then something is */
    /* missing in the data block, so we return False indicating an error. */
    while ((DataIndex < BitCheckIndex)
      || ((DataIndex == BitCheckIndex) && (Mask == 0x20)))
      {
        uchar     Temp;

        Temp = (Data[DataIndex]) - 32;
        if (Temp >= 96)
          {
            return False;  /* shouldn't ever happen! */
          }
        if (Temp < 32)
          {
            if (((Data[BitCheckIndex] - 32) & Mask) != 0)
              {
                Temp += 96; /* add in if bit was set, else leave low */
              }
            Mask = Mask >> 1;
            if (Mask == 0)
              {
                Mask = 0x20;
                BitCheckIndex -= 1;
              }
          }
        Data[DataIndex] = Temp;
        DataIndex += 1;
      }
    /* at exit:  EncodedData[DataIndex] should be FIRST resolution byte. */
    /* It's also the total number of data bytes. */
    /* and EncodedData[BitCheckIndex] should be FIRST resolution byte, OR */
    /* if Mask == 0x20, then it wrapped around falsly, and should be */
    /* LAST data byte. */
    if (Mask == 0x20)
      {
        BitCheckIndex += 1; /* now we fix the wrap-around. */
      }

    (*Length) = DataIndex; /* remote resolution bits from end */
    if (BitCheckIndex != DataIndex)
      {
        return False;  /* this means the ambiguity resolution failed */
      }

    /* now it's time to unstrech. */
    DataIndex = 0;
    FinalIndex = 0;
    while ((*Length) - DataIndex >= 7)
      {
        /* while there is still some stuff to do */
        /* the bits DataIndex+0 .. DataIndex+5 are all fine, except they are missing */
        /* the 8th bit, which is packed into DataIndex+6.  This makes for a total */
        /* of 7 bytes.  If there are less than that, then the packed bits byte is */
        /* the last one */
        for (Count = 0; Count < 6; Count += 1)
          {
            Data[FinalIndex] = Data[DataIndex + Count];
            /* oldest bit [0] is in position 5.  newest [5] is in position 0 */
            /* thus, shift %00100000 (0x20) right by offset from start. */
            if ((Data[DataIndex + 6] & (0x20 >> Count)) != 0)
              {
                Data[FinalIndex] |= 0x80; /* set the high bit */
              }
            FinalIndex += 1;
          }
        DataIndex += 6 + 1;
      }
    /* this cleans up any non multiple of 7 byte block.  If there was a */
    /* multiple of 7 bytes, (raw, not encoded) then (*Length) - DataIndex - 1 == -1, */
    /* so this loop will not be executed. */
    for (Count = 0; Count < (*Length) - DataIndex - 1; Count += 1)
      {
        Data[FinalIndex] = Data[DataIndex + Count];
        /* oldest bit [0] is in position 5.  newest [5] is in position 0 */
        /* thus, shift %00100000 (0x20) right by offset from start. */
        if ((Data[(*Length) - 1]
          & (0x20 >> (Count + 6 - ((*Length) - DataIndex - 1)))) != 0)
          {
            /* the Count + 6 - ((*Length) - DataIndex - 1) is because */
            /* if there are fewer than 6 bits, they are packed in the low */
            /* end of the byte */
            Data[FinalIndex] |= 0x80; /* set the high bit */
          }
        FinalIndex += 1;
      }
    (*Length) = FinalIndex;
    return True; /* if we got this far, then the data is probably good. */
    /* we could check to see that the last byte didn't have any extraneous */
    /* bits, but they probably wouldn't matter much anyway, and I don't */
    /* feel like doing it. */
  }


void      SendAcknowledgePacket(short PacketID)
  {
    char      Temp[MaxTotalPacketSize];
    TransRec  Packet;

    Packet.Data = Temp;  /* fudge in a data block for our fake packet. */
    Packet.Length = 2;
    Temp[0] = PacketID & 0x00ff;
    Temp[1] = (PacketID >> 8) & 0x00ff;
    ErrorEncode(&Packet);
    TransmitEncode(&Packet);
    WriteBlock(Temp,Packet.Length);
  }


void      SendPacketRequest(short PacketID)
  {
    char      Temp[MaxTotalPacketSize];
    TransRec  Packet;

    Packet.Data = Temp;  /* fudge in a data block for our fake packet. */
    Packet.Length = 3;
    Temp[0] = PacketID;
    Temp[1] = PacketID >> 8;
    Temp[2] = 0x00ff;
    ErrorEncode(&Packet);
    TransmitEncode(&Packet);
    WriteBlock(Temp,Packet.Length);
  }


short     PMod(long RawID)
  {
    while (RawID > MAXPACKETID)
      {
        RawID -= ((long)MAXPACKETID - (long)MINPACKETID + 1);
      }
    while (RawID < MINPACKETID)
      {
        RawID += ((long)MAXPACKETID - (long)MINPACKETID + 1);
      }
    return RawID;
  }


long          TickCount(void)
  {
    return TickTrack;
  }


/* here's a hack because UNIX doesn't have good PTY buffering */
void          WriteBlock(char* Data, long Length)
  {
    long    Scan;

#ifdef dBUG
    for (Scan = 0; Scan < Length; Scan += 1)
      {
        if (((unsigned char)Data[Scan] < 32) && (Data[Scan] != 0x0d))
          {
            APRINT(("Bad character in output %d\n",(int)(unsigned char)Data[Scan]));
          }
        if ((NumDataBits == 7) && ((Data[Scan] & 0x80) != 0))
          {
            APRINT(("High bit set when in 7-bit mode\n"));
          }
      }
#endif
    if ((((GlobalBufEnd - GlobalBufStart) & GlobalBufMask) + Length) < GlobalBufSize - 1)
      {
        /* there is enough room in the buffer to write data */
        while (Length > 0)
          {
            GlobalBuffer[GlobalBufEnd] = *Data;
            Data += 1;
            Length -= 1;
            GlobalBufEnd = (GlobalBufEnd + 1) & GlobalBufMask;
          }
      }
    /* else, we just don't write this packet */
  }


static long   LastPacketSend = 0;
#define LIMIT (60)


MyBoolean	CanWeSendPacket(void)
  {
    if ((signed long)TickCount() - LastPacketSend >= LIMIT)
      {
        CanTransmitFlag = True;
        LastPacketSend = TickCount();
      }
    return CanTransmitFlag;
  }
