/* CMyDocument.h */
/* This is the UNIX end of the MultiSession controller */

#ifndef includedcontrol
#define includedcontrol

#include "pty.h"

/* Generic definitions that are unavailable on UNIX */
/* typedef unsigned short ushort; */
typedef unsigned long ulong;
typedef unsigned char uchar;
#define True (1)
#define False (0)
#define NIL ((void*)0L)
typedef int MyBoolean;


#define MaxNumSessions (16)
#define BuffSize (1024) /* must be power of 2 */
#define BuffMask (BuffSize - 1)
#define MaxTransactionSize (255)
#define MaxPendingPackets (10)
#define MinPacketSize (3) /* 2 for packet ID and 1 for number of segment */
#define MaxChannelZeroCmdSize (8)
#define AbsoluteMaxPacketDataSize (2048)
#define MSModeCancelChar (24)

/* Maximum pre-transmit encoded data size is */
/*  AbsoluteMaxPacketDataSize (for data & routing info) */
/*  + 4 (for CRC) */
#define MaxTotalDataPacketSizeLocalToHere (AbsoluteMaxPacketDataSize + 4)
/* That needs encoding: for 8 bits */
/*  MaxTotalDataPacketSizeLocalToHere (for data) */
/*  + MaxTotalDataPacketSizeLocalToHere / 7 + 1  (for resolution bits) */
/*  + 1  (for terminator) */
/* For 7 bits: */
/*  MaxTotalDataPacketSizeLocalToHere * 8 / 7 + 1  (for streched data) */
/*  + MaxTotalDataPacketSizeLocalToHere / 6 + 1  (for resolution bits) */
/*  + 1  (for terminator) */
/* 7 bit encoding will be bigger, so we'll use that */
#define MaxTotalPacketSize (((MaxTotalDataPacketSizeLocalToHere * 8) / 7) + 1 \
  + (MaxTotalDataPacketSizeLocalToHere / 6) + 1 \
  + 1 /* terminator */)

#define READBUFSIZE (MaxTotalPacketSize*3/2)

typedef struct BuffRec
  {
    short     DataStart;
    short     DataEnd;
    char      Data[BuffSize];
  } BuffRec;

typedef struct TransRec
  {
    /* pointer to buffer of size MaxTotalPacketSize */
    char*       Data;
    /* number of valid bytes in buffer, -1 == packet doesn't exist (not received/sent) */
    short       Length;
    /* for SendPackets, indicates packet has been acknowledged */
    MyBoolean   Acknowledged;
    /* for SendPackets, this is the last time the packet was sent so we can automatically */
    /* resend the packet when the timeout period expires. */
    long        LastActionTime;
  } TransRec;


#ifndef CompilingCMyDocument

/* high level routing of data */
extern CTerminal*   Handlers[MaxNumSessions]; /* NIL == not in use */
extern BuffRec      Buffers[MaxNumSessions];
extern char         MakeBuffers[MaxNumSessions][AbsoluteMaxPacketDataSize];

/* low level lost packet replacement things */
extern TransRec     ReceivePackets[MaxPendingPackets];
extern TransRec     SendPackets[MaxPendingPackets];
extern short        ReceiveBasePacketIDNum;
extern short        ReceiveNumPackets;
extern short        SendBasePacketIDNum;
extern short        SendNumPackets;

/* Session 0 is special.  Its CTerminal Handler is the root console, but it */
/* does not receive data unless DataMode == SingleSession, in which case it */
/* sends raw data.  Data received from session 0 is for packet validation and */
/* commands.  The other sessions are normal */
extern char         PartialPacketArray[READBUFSIZE]; /* DoIdle stores partial packets here */
extern short        PartialPacketLength;
extern long         TimeOutFactor;
extern char         ChannelZeroBuffer[MaxChannelZeroCmdSize];
extern short        ChannelZeroBufPtr;
extern short        EffectiveMaxPacketSize;

extern short        NumDataBits;

extern volatile MyBoolean   QuitFlag;
extern volatile long  TickTrack;

#endif


void      InitializeCMyDocument(void);
void      ShutDownCMyDocument(void);
void      GoAway(void);

MyBoolean	ReceiveDataFromLocalTerminal(char* Data, short Length, CTerminal* Sender);
MyBoolean	PostPacket(char* Data, short Length, short SessionNumber);
void      DocumentIdle(void);
void      ReceiveChannelZeroCommand(char* Data, short Length);
void      DoNewSessionRequest(void);
void      DoNewSessionGranted(short SessionID);
void      DoCloseSessionCommand(short SessionID);
void      SendChannelZeroCommand(char* Data, short Length);
void      SendAcknowledgePacket(short PacketID);
void      SendPacketRequest(short PacketID);
void      MakePacket(short PacketID, TransRec* Packet); /* construct packet from buffers */
void      UnmakePacket(char* Data, short Length); /* dispatch packet to handlers */
void      ErrorEncode(TransRec* Packet);
MyBoolean	ErrorDecode(char* Data, short* Length);
void      TransmitEncode(TransRec* Packet);
void      TransmitEncode8(TransRec* Packet);
void      TransmitEncode7(TransRec* Packet);
MyBoolean	TransmitDecode(char* Data, short* Length);
MyBoolean	TransmitDecode8(char* Data, short* Length);
MyBoolean	TransmitDecode7(char* Data, short* Length);
void      TerminalWantsToDieCloseSession(CTerminal* OneThatDied);
MyBoolean	CheckPacketFormat(char* Data, short Length);
void      WriteBlock(char* Data, long Length);
MyBoolean	CanWeSendPacket(void);
void      DoAcknowledgePacket(short PacketID);
short     PMod(long RawID);

long      TickCount(void);

#endif
