/* main.c */

#define compilingmain

#include "control.h"

#define RATE (3L) /* 60ths of a second */

static jmp_buf  sjbuf;

#define SleepTime (3)

extern int errno;


volatile short    Idling = False;

void    IdleProc(void)
  {
    TickTrack += RATE;
    if (!Idling)
      {
        /* if we are not handling an idle event currently, we can enter */
        /* one now. */
        Idling = True;
        signal(SIGALRM,(void*)&IdleProc);
        PtyIdle();
        DocumentIdle();
        PtyIdle();
        Idling = False;
        longjmp(sjbuf,0); /* restart */
      }
     else
      {
        /* if we are handling an idle event currently, we increment */
        /* the counter, but we don't actually call the procedures */
        signal(SIGALRM,(void*)&IdleProc);
        return; /* return to location of interrupt */
      }
  }


struct utmp* RootUtmpEntry; /* referenced in pty.c */

void    main(int argc, char* argv[])
  {
    long              LastTime;
    struct itimerval  Timer;
    struct itimerval  TimerOld;
    TTYInfo           NewTTY;
    char*             RootTTYName;
    int               TTYDesc;
    int               RootUtmpIndex;

    printf("\nMultiSession Server V1.0.2, by Thomas R. Lawrence\n\n");
    fflush(stdout);
#ifdef dBUG
    bugfile = fopen("audit","w");
    fprintf(bugfile,"Debugging:\n");
    fflush(bugfile);
#endif
    TTYDesc = open("/dev/tty",O_RDWR);
    NumDataBits = 8;
    if (argc == 2)
      {
        NumDataBits = argv[1][0] - '0';
        if ((NumDataBits != 7) && (NumDataBits != 8))
          {
            char  Msg[] = "Number of data bits must be 7 or 8\n";

            write(1,Msg,strlen(Msg));
            exit(0);
          }
      }
    write(1,"\x18Kajim",6); /* turn on Macintosh MultiSession mode */

    OpenUtmpFile();
    RootTTYName = (char*)ttyname(TTYDesc);
    if (RootTTYName == NIL)
      {
        printf("MultiSession requires a pseudo-TTY to run.");
        exit(0);
      }
    RootUtmpIndex = FindUtmpEntry(RootTTYName);
    GetUtmpEntry(RootUtmpIndex,RootUtmpEntry);
    DeleteUtmpEntry(RootUtmpIndex);

    GetTTYInfo(0,&OriginalTTY);
    NewTTY = OriginalTTY;
    MakeSpiffyTTY(&NewTTY);
    SetTTYInfo(0,&NewTTY);

    signal(SIGQUIT,SIG_IGN);

    InitializeCMyDocument();
    LastTime = TickCount();
    Timer.it_interval.tv_sec = 0;
    Timer.it_interval.tv_usec = RATE * 16666L;
    Timer.it_value.tv_sec = 0;
    Timer.it_value.tv_usec = 10000; /* start with .1 second to go */
    TimerOld = Timer;
    setitimer(ITIMER_REAL,&Timer,&TimerOld);
    setjmp(sjbuf);
    signal(SIGALRM,(void*)&IdleProc);
    while (!QuitFlag)
      {
        pause();
      }
    signal(SIGALRM,SIG_IGN);
    ShutDownCMyDocument();

    /* restore the original login */
    PutUtmpEntry(RootUtmpIndex,RootUtmpEntry);

    /* try to turn off the remote system */
    /* it expects 5, we'll send 10 */
    {
      char      TurnOffMsg[] = {MSModeCancelChar,MSModeCancelChar,
        MSModeCancelChar,MSModeCancelChar,MSModeCancelChar,MSModeCancelChar,
        MSModeCancelChar,MSModeCancelChar,MSModeCancelChar,MSModeCancelChar};

        write(1,TurnOffMsg,10);
    }

    /* wait for quiescence on input port */
    {
      char    Buffer[256];
      int     Length;
      int     CycleCount;

      CycleCount = 0;
      while (CycleCount < 3)
        {
          Length = read(0,Buffer,256);
          if (Length > 0)
            {
              CycleCount = 0;
            }
          sleep(1);
          CycleCount += 1;
        }
      }
    SetTTYInfo(0,&OriginalTTY);
    {
      char    OtherMessage[] = "\n\nDone.\n\n";

      write(0,OtherMessage,strlen(OtherMessage));
    }
  }
