/* pty.c */

#include "pty.h"
#include "control.h"


TTYInfo     OriginalTTY;

#ifdef dBUG
FILE*       bugfile;
#endif

char*       SHELLNAME = "/bin/csh";

extern int  errno;

static CTerminal*   Root = NIL;


/* Close a PTY access channel and kill the process running on it. */
void          DeletePty(CTerminal* TheTerminal)
  {
    CTerminal*    Scan;
    CTerminal*    Lag;
    int           error;
    short         SessionNum;

    APRINT(("+DeletePty\n"));
    /* removing login entry */
    DeleteUtmpEntry(TheTerminal->UtmpIndex);
    /* closing file descriptor */
    if (TheTerminal->PID != 0)
      {
        int         Result;

        Result = kill(TheTerminal->PID,SIGKILL); /* unconditionally terminate the process */
      }
    error = close(TheTerminal->MasterDesc);

    while (TheTerminal->OldestBuffer != NIL)
      {
        PtyBufRec*    Temp;

        Temp = TheTerminal->OldestBuffer;
        TheTerminal->OldestBuffer = (TheTerminal->OldestBuffer)->NewerNext;
        free (Temp);
      }

    /* removing entry from linked list */
    Scan = Root;
    Lag = NIL;
    while ((Scan != TheTerminal) && (Scan != NIL))
      {
        Lag = Scan;
        Scan = Scan->Next;
      }
    ERROR(Scan==NIL,{APRINT(("fatal internal linked list error\n"));exit(1);})
    SessionNum = 0;
    while ((SessionNum < MaxNumSessions) && (Handlers[SessionNum] != TheTerminal))
      {
        SessionNum += 1;
      }
    ERROR(SessionNum>=MaxNumSessions,{APRINT(("fatal internal Handlers table error\n"));exit(1);})
    Handlers[SessionNum] = NIL;

    if (Lag == NIL)
      {
        Root = TheTerminal->Next;
      }
     else
      {
        Lag->Next = TheTerminal->Next;
      }
    free(TheTerminal);
    APRINT(("-DeletePty\n"));
  }


/* Attempts to create a new PTY.  Returns a pointer to it's memory cell if */
/* successful, or returns NIL if it failed. */
CTerminal*    NewPty(void)
  {
    char        Name[] = "/dev/pty..";
    char        Alpha[] = "pqrstuvwxyzPQRST"; /* pooh */
    char        Hex[16] = "0123456789abcdef";
    short       AlphaScan;
    short       NumberScan;
    int         MasterSideDesc;
    int         SlaveSideDesc;
    CTerminal*  Ref;
    int         PID;
    TTYInfo     NewTTY;
    int         error;

    APRINT(("+NewPty\n"));
    /* first, we try to find a PTY that's available. */
    for (AlphaScan = 0; AlphaScan < 16; AlphaScan += 1)
      {
        for (NumberScan = 0; NumberScan < 16; NumberScan += 1)
          {
            Name[8] = Alpha[AlphaScan];
            Name[9] = Hex[NumberScan];
            MasterSideDesc = open(Name,O_RDWR /*| O_NOCTTY*/ | O_NDELAY,0);
            if (MasterSideDesc >= 0) /* no error */
              {
                Name[5] = 't'; /* try to open the slave side, with blocking */
                /* should we ask for exclusive use here? */
                SlaveSideDesc = open(Name,O_RDWR /*| O_NOCTTY*/,0);
                if (SlaveSideDesc >= 0) /* no error */
                  {
                    goto WeHavePtyPoint;
                  }
                Name[5] = 'p';
              }
          }
      }
    APRINT(("-NewPty failed, no openable PTYs\n"));
    return NIL;  /* return NIL if we failed to open a PTY */

   WeHavePtyPoint:
    /* now we have to configure it */
    /* first, closing slave side */
    error = close(SlaveSideDesc);
    /* Now, setting the modes for the server's side (master) */
    NewTTY = OriginalTTY;
    MakeSpiffyTTY(&NewTTY);
    SetTTYInfo(MasterSideDesc,&NewTTY);
    /* now, trying to make us exclusive owners of the TTY */
/*
    error = seteuid(getuid(),geteuid());
    error = chown(Name,getuid(),getgid());
    ERROR(error==-1,APRINT(("chown on slave TTY failed %d\n",errno)));
    error = chmod(Name,0600);
    ERROR(error==-1,APRINT(("chmod on slave TTY failed %d\n",errno)));
*/

    /* now we start the new process */
    PID = fork();
    if (PID == -1)
      {
        /* if we were unable to create the process, then we fail. */
        error = close(MasterSideDesc);
        APRINT(("-NewPty fork failed\n"));
        return NIL;
      }
    if (PID != 0)
      {
        /* master side */
#if 0
        error = setreuid(geteuid(),geteuid());
#endif
        Ref = (void*) malloc(sizeof(CTerminal));
        Ref->Next = Root;
        Root = Ref;
        Ref->MasterDesc = MasterSideDesc; /* remember file descriptor */
        Ref->PID = PID; /* remember process ID */
        Ref->OldestBuffer = NIL;
        Ref->NewestBuffer = NIL;
        Ref->UtmpIndex = MakeUtmpEntry(Name);
        APRINT(("-NewPty master side, success\n"));
        return Ref;
      }
     else
      {
        int                 error;
        int                 CountieThingie;
        int                 MyPid;
        struct itimerval    Timer;
        struct itimerval    TimerOld;
        char*               stupidparameters[2];
        char*               shell;

        /* slave side */
        close(MasterSideDesc); /* ya wanna make somethin' of it? */
        close(0);
        close(1);
        close(2);

#if 0
        /* I have no idea why this is done, but everyone else does it. */
        setuid(getuid());
        setgid(getgid());
#endif

        MyPid = getpid();
        ioctl(0,TIOCSPGRP,&MyPid); /* what does this do? */
        setpgrp(0,MyPid);

        /* I sure hope when we open the TTY, it becomes our */
        /* controlling terminal */
        SlaveSideDesc = open(Name,O_RDWR /*| O_NOCTTY*/,0);
        ioctl(SlaveSideDesc,TIOCEXCL,(void*)NULL);

        SetTTYInfo(SlaveSideDesc,&OriginalTTY);

        /* install the terminal in our standard streams */
        /* the open installed the terminal into 0 */
        error = dup(SlaveSideDesc); /* install into 1 */
        error = dup(SlaveSideDesc); /* install into 2 */

        ioctl(open("/dev/tty",O_RDWR),TIOCNOTTY,(char*)0);
        for (CountieThingie = getdtablesize() - 1; CountieThingie > 2; CountieThingie -= 1)
          {
            close(CountieThingie);
          }

        /* disabling timer */
        Timer.it_interval.tv_sec = 0;
        Timer.it_interval.tv_usec = 0;
        Timer.it_value.tv_sec = 0;
        Timer.it_value.tv_usec = 0;
        setitimer(ITIMER_REAL,&Timer,&TimerOld);

        signal(SIGALRM,SIG_DFL); /* restoring default signals */
        signal(SIGQUIT,SIG_DFL);

        /* shelling out to the shell in the shell variable or the default shell */
        /* in SHELLNAME if there is no shell variable */
        shell = getenv("SHELL");
        if (shell == NIL)
          {
            shell = SHELLNAME;
          }
        stupidparameters[0] = shell; /* shell */
        stupidparameters[1] = NIL;
        APRINT(("execv'ing on slave side\n"));
        execv(SHELLNAME,stupidparameters);  /* bring up a shell */
        APRINT(("execv on slave side failed %d\n",errno));
        exit(1); /* abort, should never happen, but it might, nahh it couldn't... */
      }
  }


/* Send data to a PTY */
void          DispatchPty(CTerminal* TheTerminal, char* TheData, long NumBytes)
  {
    int           error;
    PtyBufRec*    Temp;

    while (NumBytes > 0)
      {
        /* making and linking buffers onto the output queue */
        Temp = (void*) malloc(sizeof(PtyBufRec));
        if (NumBytes > PtyBufSize)
          {
            Temp->NumBytesInBuffer = PtyBufSize;
          }
         else
          {
            Temp->NumBytesInBuffer = NumBytes;
          }
        memcpy(&(Temp->Data[0]),TheData,Temp->NumBytesInBuffer);
        /* establising link */
        Temp->NewerNext = NIL;
        if (TheTerminal->NewestBuffer == NIL)
          {
            TheTerminal->NewestBuffer = Temp;
            TheTerminal->OldestBuffer = Temp;
          }
         else
          {
            TheTerminal->NewestBuffer->NewerNext = Temp;
            TheTerminal->NewestBuffer = Temp;
          }
        TheData += Temp->NumBytesInBuffer;
        NumBytes -= Temp->NumBytesInBuffer;
      }
  }


/* Read data from PTYs.  Also detect terminated processes. */
void          PtyIdle(void)
  {
    CTerminal*    Scan;
    char          Buffer[256];
    int           NumBytes;
    short         Item;
    pid_t         PID;
    union wait    Status;

    Scan = Root;
    while (Scan != NIL)
      {
        Item = 0;
        /* identifying which session it is */
        while (Handlers[Item] != Scan)
          {
            Item += 1;
          }
        /* seeing if there is any data to read */
        if (((Buffers[Item].DataEnd - Buffers[Item].DataStart) & BuffMask)
          + 255 < BuffSize - 1)
          {
            /* there is room to write data to the local buffer. */
            NumBytes = read(Scan->MasterDesc,Buffer,255);
            if (NumBytes > 0)
              {
                /* some data was read without error, we can send it. */
                ReceiveDataFromLocalTerminal(Buffer,NumBytes,Scan);
              }
          } /* end of if there's room in our local buffer */
        /* seeing if there is any data to write */
        if (Scan->OldestBuffer != NIL)
          {
            NumBytes = write(Scan->MasterDesc,Scan->OldestBuffer->Data,
              Scan->OldestBuffer->NumBytesInBuffer);
            if (NumBytes > 0)
              {
                /* if data was successfully written, see how much */
                if (NumBytes == Scan->OldestBuffer->NumBytesInBuffer)
                  {
                    PtyBufRec*    Temp;

                    /* entire write was successful, we can dequeue */
                    Temp = Scan->OldestBuffer;
                    Scan->OldestBuffer = Scan->OldestBuffer->NewerNext;
                    if (Scan->OldestBuffer == NIL)
                      {
                        /* this happens when there was only one buffer */
                        Scan->NewestBuffer = NIL;
                      }
                    free(Temp);
                  }
                 else
                  {
                    int     Inxs;

                    /* the entire write was not successful, but some was */
                    for (Inxs = 0; Inxs < Scan->OldestBuffer->NumBytesInBuffer - NumBytes;
                      Inxs += 1)
                      {
                        Scan->OldestBuffer->Data[Inxs]
                          = Scan->OldestBuffer->Data[NumBytes + Inxs];
                      }
                    Scan->OldestBuffer->NumBytesInBuffer -= NumBytes;
                  }
              }
          } /* end of writing to the PTY */
        Scan = (*Scan).Next;
      } /* end of scan through all pty loop */
    PID = wait3(&Status,WNOHANG,NIL/*no usage*/);
    if (PID > 0)
      {
        CTerminal*    Check;

        Check = Root;
        while (Check != NIL)
          {
            if (Check->PID == PID)
              {
                if ((WIFEXITED(Status)) || WIFSIGNALED(Status))
                  {
                    /* the process exited, so don't kill it */
                    /* We don't do this if it was only WIFSTOPPED */
                    /* since we have to kill stopped processes ourself */
                    Check->PID = 0;
                  }
                TerminalWantsToDieCloseSession(Check);
                Check = NIL;
              }
             else
              {
                Check = (*Check).Next;
              }
          }
        /* if the process's record doesn't exist, then */
        /* we just fall through the loop.  this happens if it was killed */
        /* but not waited for elsewhere. */
      } /* end of if wait3's pid != 0 */
  }


void          WindowSizeChange(CTerminal* TheTerminal, short NewX, short NewY)
  {
    struct winsize    MyWindowSize;

    APRINT(("+WindowSizeChange\n"));
#ifdef TIOCSWINSZ
    MyWindowSize.ws_col = NewX;
    MyWindowSize.ws_row = NewY;
    MyWindowSize.ws_xpixel = 0;
    MyWindowSize.ws_ypixel = 0;
    ioctl(TheTerminal->MasterDesc,TIOCSWINSZ,&MyWindowSize);
#endif
    APRINT(("-WindowSizeChange\n"));
  }


/****************************************************************/
/* Getting and Setting and Getting and Setting and Getting TTY Stuff */


void    GetTTYInfo(int TheFile, TTYInfo* TheInfo)
  {
    APRINT(("+GetTTYInfo\n"));
    ioctl(TheFile,TIOCGETC,&(TheInfo->TCHARS));
    ioctl(TheFile,TIOCGETD,&(TheInfo->LDISC));
    ioctl(TheFile,TIOCGETP,&(TheInfo->TTYB));
    ioctl(TheFile,TIOCGLTC,&(TheInfo->LTCHARS));
    ioctl(TheFile,TIOCLGET,&(TheInfo->LMODE));
    TheInfo->fcntlFlags = fcntl(TheFile,F_GETFL,0);
    APRINT(("-GetTTYInfo\n"));
  }


void    SetTTYInfo(int TheFile, TTYInfo* TheInfo)
  {
    APRINT(("+SetTTYInfo\n"));
    ioctl(TheFile,TIOCSETC,&(TheInfo->TCHARS));
    ioctl(TheFile,TIOCSETD,&(TheInfo->LDISC));
    ioctl(TheFile,TIOCSETP,&(TheInfo->TTYB));
    ioctl(TheFile,TIOCSLTC,&(TheInfo->LTCHARS));
    ioctl(TheFile,TIOCLSET,&(TheInfo->LMODE));
    fcntl(TheFile,F_SETFL,TheInfo->fcntlFlags);
/*
    error = ioctl(TheFile,TIOCEXCL,0);
    ERROR(error==-1,APRINT(("SetTTYInfo ioctl for exclusive use error %d\n",errno)));
*/
    APRINT(("-SetTTYInfo\n"));
  }


void    MakeSpiffyTTY(TTYInfo* TheInfo)
  {
    APRINT(("+MakeSpiffyTTY\n"));
    TheInfo->TTYB.sg_flags = RAW | ANYP;
    TheInfo->TCHARS.t_intrc = -1;
    TheInfo->TCHARS.t_quitc = -1;
    TheInfo->TCHARS.t_startc = -1;
    TheInfo->TCHARS.t_stopc = -1;
    TheInfo->LTCHARS.t_suspc = -1;
    TheInfo->LTCHARS.t_dsuspc = -1;
    TheInfo->LTCHARS.t_flushc = -1;
    TheInfo->LTCHARS.t_lnextc = -1;
    TheInfo->fcntlFlags |= O_NDELAY;
    APRINT(("-MakeSpiffyTTY\n"));
  }


/****************************************************************/
/* Fun with UTMP! */


static int      UtmpFile;
static short    UtmpValid = False;


void    OpenUtmpFile(void)
  {
    APRINT(("+OpenUtmpFile\n"));
#ifdef USE_UTMP
    UtmpFile = open("/etc/utmp",O_RDWR,0);
    if (UtmpFile != -1)
      {
        UtmpValid = True; /* opened successfully, so we can use it */
      }
#endif
    APRINT(("-OpenUtmpFile\n"));
  }


/* return the index number of the TTYName in the TTY list. */
/* 0 = not found, 1 = first one, 2 = second one... */
int     FindUtmpEntry(char* TTYName)
  {
    struct ttyent*    Thingie;
    char*             SpiffName;
    int               Index;

    APRINT(("+FindUtmpEntry\n"));
#ifdef USE_UTMP
    if (UtmpValid)
      {
        SpiffName = (char*)rindex(TTYName,'/');
        if (SpiffName != NIL)
          {
            SpiffName += 1; /* step past '/' to first character of name */
          }
         else
          {
            SpiffName = TTYName;
          }
        setttyent(); /* reset TTY scanning thing */
        Index = 0;
       Loop:
        Index += 1;
        Thingie = getttyent(); /* get a tty thing */
        if (Thingie == NIL)
          {
            APRINT(("-FindUtmpEntry found nothing\n"));
            return 0; /* means we found nothing */
          }
        if (strcmp(Thingie->ty_name,SpiffName) == 0)
          {
            APRINT(("-FindUtmpEntry %d\n",(int)Index));
            /* names are equal, so we've found our man */
            return Index;
          }
        goto Loop;
      }
#endif
    APRINT(("-FindUtmpEntry found nothing\n"));
    return 0; /* this gets returned if we couldn't access file */
  }


int     GetUtmpEntry(int Index, struct utmp* TheUtmpEntry)
  {
    int     BytesRead = 0;

    APRINT(("+GetUtmpEntry\n"));
#ifdef USE_UTMP
    if (UtmpValid)
      {
        /* first, go to the location in UTMP */
        lseek(UtmpFile,(long)(Index * sizeof(struct utmp)),0);
        BytesRead = read(UtmpFile,(void*)TheUtmpEntry,sizeof(struct utmp));
      }
#endif
    APRINT(("-GetUtmpEntry %d\n",(int)BytesRead));
    return (BytesRead > 0); /* true = worked, false = failed */
  }


void    PutUtmpEntry(int Index, struct utmp* TheUtmpEntry)
  {
    int     BytesWritten;

    APRINT(("+PutUtmpEntry\n"));
#ifdef USE_UTMP
    if (UtmpValid)
      {
        lseek(UtmpFile,(long)(Index * sizeof(struct utmp)),0);
        BytesWritten = write(UtmpFile,(void*)TheUtmpEntry,sizeof(struct utmp));
      }
#endif
    APRINT(("-PutUtmpEntry\n"));
  }


void    DeleteUtmpEntry(int Index)
  {
    struct utmp   ZapEntry;
    int           Scan;

    APRINT(("+DeleteUtmpEntry\n"));
#ifdef USE_UTMP
    for (Scan = 0; Scan < sizeof(struct utmp); Scan += 1)
      {
        /* portable? nahhh.. */
        ((char*)(&ZapEntry))[Scan] = 0;
      }
    PutUtmpEntry(Index,&ZapEntry);
#endif
    APRINT(("-DeleteUtmpEntry\n"));
  }


extern struct utmp* RootUtmpEntry; /* from main.c */

int     MakeUtmpEntry(char* Name)
  {
    struct utmp   OurUtmpEntry;
    int           Where;

    APRINT(("+MakeUtmpEntry\n"));
#ifdef USE_UTMP
    Where = FindUtmpEntry(Name);
    if (UtmpValid && (Where != 0))
      {
        if (rindex(Name,'/') == 0)
          {
            strncpy(OurUtmpEntry.ut_line,Name,8);
          }
         else
          {
            strncpy(OurUtmpEntry.ut_line,(rindex(Name,'/') + (char*)1),8);
          }
        strncpy(OurUtmpEntry.ut_name,RootUtmpEntry->ut_name,8);
        strncpy(OurUtmpEntry.ut_host,"Unknown",8);
        time(&OurUtmpEntry.ut_time);
        PutUtmpEntry(Where,&OurUtmpEntry);
        APRINT(("-MakeUtmpEntry %d\n",(int)Where));
        return Where;
      }
#endif
    APRINT(("-MakeUtmpEntry 0\n"));
    return 0;
  }


/****************************************************************/


#ifdef dBUG
void    AuditPrint(char* First, ...)
  {
    va_list     ap;

    va_start(ap,First);
    vfprintf(bugfile,First,ap);
    va_end(ap);
    fflush(bugfile);
  }
#endif
