/* pty.h */
/* Pseudo-terminal handler */

#ifndef includedpty
#define includedpty

/* comment out this line to turn off audit file generation */
/* #define dBUG */

/* debugging macro */
#ifdef dBUG
#define APRINT(param) AuditPrint param
#define ERROR(cond,cmd) if (cond) cmd
void    AuditPrint(char* First, ...);
#else
#define APRINT(param)
#define ERROR(cond,cmd)
#endif

/* if you want to try to modify the UTMP file, define this macro. */
/* WARNING:  I have no idea if it works or not! */
/* #define USE_UTMP */

#include <stdio.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <signal.h>
#include <sys/time.h>
#include <setjmp.h>
#include <utmp.h>
#include <ttyent.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <stdarg.h>


#define PtyBufSize (64)
typedef struct Buf
  {
    struct Buf*   NewerNext;
    short         NumBytesInBuffer;
    char          Data[PtyBufSize];
  } PtyBufRec;

typedef struct CT
  {
    int           MasterDesc;  /* file reference number */
    int           PID;
    int           UtmpIndex;
    struct CT*    Next; /* internal link, not for use outside of "pty.c" */
    PtyBufRec*    OldestBuffer; /* NIL = none */
    PtyBufRec*    NewestBuffer; /* NIL = none */
  } CTerminal;


void          DeletePty(CTerminal* TheTerminal);
CTerminal*    NewPty(void); /* returns NIL if pty creation failed */
void          DispatchPty(CTerminal* TheTerminal, char* TheData, long NumBytes);
void          PtyIdle(void);
void          WindowSizeChange(CTerminal* TheTerminal, short NewX, short NewY);


typedef struct
  {
    struct sgttyb     TTYB;
    struct tchars     TCHARS;
    struct ltchars    LTCHARS;
    int               LDISC;
    int               LMODE;
    long              fcntlFlags;
  } TTYInfo;

extern TTYInfo    OriginalTTY;
#ifdef dBUG
extern FILE*      bugfile;
#endif

/* tty stuffies */
void      GetTTYInfo(int TheFile, TTYInfo* TheInfo);
void      SetTTYInfo(int TheFile, TTYInfo* TheInfo);
void      MakeSpiffyTTY(TTYInfo* TheInfo);

/* utmp thingies */
void      OpenUtmpFile(void);
int       FindUtmpEntry(char* TTYName);
int       GetUtmpEntry(int Index, struct utmp* TheUtmpEntry);
void      PutUtmpEntry(int Index, struct utmp* TheUtmpEntry);
void      DeleteUtmpEntry(int Index);

#endif
