/* ReceiveFile.c */

#include "commonstuff.h"

enum
  {
    ReceiveDoingNothing, /* nothing interesting going on */
    ReceiveWaitingForHeader, /* waiting for file header to be sent (we'll ack when we get it) */
    WaitingForHeader_Acking, /* acking, waiting for first data block */
    ReceivingData, /* receiving data, waiting for "all done with file", request resends */
    ReceivingData_Cleanup, /* "all done with file" gotten, correcting errors */
    SendGoodWaitForHeader, /* sending "all done", waiting for next header or quit */
    ReceiveGoAway /* when last packet is send, then go away */
  };


short           ReceiveBufPtr;
short           SendBufPtr;
short           CurrentFileDescriptor;
long            CurrentFileLength;
CSack*          BadBlockList; /* list of 1K idices of all corrupted blocks */
CSack*          OtherBadBlockList;
short           State;
unsigned long   LastTimeOut;
short           NumDataBits;
long            PreviousBlockIndex;
char            ReceiveBuffer[MaxTotalPacketSize];
char            SendBuffer[MaxTotalPacketSize];


void      InitModule(int argc, char* argv[]);
void      DoIdle(void);
void      ReceiveBytes(char* Data, long Length);
void      HandlePacket(char* Data, short Length);
void      CancelTransfer(void);



void        InitModule(int argc, char* argv[])
  {
    APRINT(("+InitModule\n"));

    fprintf(stdout,"Ready to receive files...");
    fflush(stdout);

    ReceiveBufPtr = 0;
    SendBufPtr = 0;
    BadBlockList = NewCSack();
    OtherBadBlockList = NewCSack();
    NumDataBits = 0; /* we don't know how many data bits are being used yet */

    State = ReceiveWaitingForHeader;
    LastTimeOut = TickCount() - RetryIntersticeDelay;

    APRINT(("-InitModule\n"));
  }


#define LocalBufSize (255)

void      DoIdle(void)
  {
    TransRec      FakePacket;
    long          BadOffset;
    long          FileBytes;
    char          LocalReadBuffer[LocalBufSize];

   TryAgainPoint:
    /* first, we see if there is any data to read */
    FileBytes = LocalBufSize;
    FileBytes = read(0,LocalReadBuffer,FileBytes);
    if (FileBytes < 0)
      {
        FileBytes = 0;
      }
    if (FileBytes > 0)
      {
        ReceiveBytes(LocalReadBuffer,FileBytes);
        goto TryAgainPoint;
      }

    /* attempting to resend some data */
    if (SendBufPtr != 0)
      {
        long        NumBytesWritten;

        NumBytesWritten = write(1,SendBuffer,SendBufPtr);
        if (NumBytesWritten <= 0)
          {
            return; /* failed altogether */
          }
        if (NumBytesWritten < SendBufPtr)
          {
            long      Scan;

            SendBufPtr -= NumBytesWritten;
            for (Scan = 0; Scan < SendBufPtr; Scan += 1)
              {
                SendBuffer[Scan] = SendBuffer[Scan + NumBytesWritten];
              }
            return;
          }
         else
          {
            SendBufPtr = 0;
          }
      }

    switch (State)
      {
        case ReceiveDoingNothing: /* nothing interesting going on */
          break;
        case ReceiveWaitingForHeader: /* waiting for file header to be sent (we'll ack when we get it) */
          break;
        case WaitingForHeader_Acking: /* acking, waiting for first data block */
          if (TickCount() - LastTimeOut >= RetryIntersticeDelay)
            {
              APRINT(("Waiting for header: acknowledging\n"));
              LastTimeOut = TickCount();
              FakePacket.Data = SendBuffer;
              FakePacket.Length = GotHeaderStartSendingLength;
              SendBuffer[0] = GotHeaderStartSendingWord;
              ErrorEncode(&FakePacket);
              TransmitEncode(&FakePacket,NumDataBits);
              SendBufPtr = FakePacket.Length;
              goto TryAgainPoint;
            }
          break;
        case ReceivingData: /* receiving data, waiting for "all done with file", request resends */
          if (NumElements(OtherBadBlockList) != 0)
            {
              /* we don't want to do anything unless there are bad packets to */
              /* resend while we are getting sequenced data */
              ResetScan(OtherBadBlockList);
              GetNext(OtherBadBlockList,&BadOffset);
              KillElement(OtherBadBlockList,&BadOffset);
              APRINT(("ReceivingData: requesting %d\n",BadOffset));
              goto SkipIntoReceivePoint;
              /* (this also avoids the timeout counter thing) */
            }
          break;
        case ReceivingData_Cleanup: /* "all done with file" gotten, correcting errors */
          if (NumElements(BadBlockList) != 0)
            {
              /* still some bad blocks to get */
              APRINT(("ReceivingData_Cleanup: still some bad blocks\n"));
             ReceivingDataStuffThingiesPoint:
              if (TickCount() - LastTimeOut >= RetryIntersticeDelay)
                {
                  LastTimeOut = TickCount();
                  ResetScan(BadBlockList);
                  GetNext(BadBlockList,&BadOffset);
                 SkipIntoReceivePoint:
                  BadOffset = BadOffset / FileXferPacketDataSize;
                  APRINT(("Requesting packet %d\n",BadOffset));
                  FakePacket.Data = SendBuffer;
                  FakePacket.Length = ResendRequestLength;
                  SendBuffer[0] = ResendRequestWord;
                  SendBuffer[ResendRequest1KIndex + 0] = BadOffset & 0x00ff;
                  SendBuffer[ResendRequest1KIndex + 1] = (BadOffset >> 8) & 0x00ff;
                  SendBuffer[ResendRequest1KIndex + 2] = (BadOffset >> 16) & 0x00ff;
                  ErrorEncode(&FakePacket);
                  TransmitEncode(&FakePacket,NumDataBits);
                  SendBufPtr = FakePacket.Length;
                  goto TryAgainPoint;
                }
            }
           else
            {
              APRINT(("ReceivingData_Cleanup: no more bad packets, waiting for next header\n"));
              State = SendGoodWaitForHeader;
              LastTimeOut = TickCount() - RetryIntersticeDelay;
              close(CurrentFileDescriptor);
              goto TryAgainPoint;
            }
          break;
        case SendGoodWaitForHeader: /* sending "all done", waiting for next header or quit */
          /* all done, sending "good" */
          if (TickCount() - LastTimeOut >= RetryIntersticeDelay)
            {
              APRINT(("SendGoodWaitForHeader: sending 'all done'\n"));
              LastTimeOut = TickCount();
              FakePacket.Data = SendBuffer;
              FakePacket.Length = GoodGotoNextFileLength;
              SendBuffer[0] = GoodGotoNextFileWord;
              ErrorEncode(&FakePacket);
              TransmitEncode(&FakePacket,NumDataBits);
              SendBufPtr = FakePacket.Length;
              goto TryAgainPoint;
            }
          break;
        case ReceiveGoAway:
          QuitFlag = True;
          APRINT(("Going away\n"));
          return;
        default:
          APRINT(("Unknown state\n"));
          exit(1);
      }
  }


void      ReceiveBytes(char* Data, long Length)
  {
    short       AccumulatedAborts;
    short       MaxAborts;
    short       DataScan;

    AccumulatedAborts = 0;
    MaxAborts = 0;
    for (DataScan = 0; DataScan < ReceiveBufPtr; DataScan += 1)
      {
        if (ReceiveBuffer[DataScan] == CANCELCHARACTER)
          {
            AccumulatedAborts += 1;
          }
         else
          {
            AccumulatedAborts = 0;
          }
        if (AccumulatedAborts > MaxAborts)
          {
            MaxAborts = AccumulatedAborts;
          }
      }
    for (DataScan = 0; DataScan < Length; DataScan += 1)
      {
        if (Data[DataScan] == CANCELCHARACTER)
          {
            AccumulatedAborts += 1;
          }
         else
          {
            AccumulatedAborts = 0;
          }
        if (AccumulatedAborts > MaxAborts)
          {
            MaxAborts = AccumulatedAborts;
          }
      }
    if (MaxAborts > ABORTTHRESHHOLD)
      {
        if (State == SendGoodWaitForHeader)
          {
            State = ReceiveGoAway;
          }
         else
          {
            APRINT(("Cancel message received, cancelling\n"));
            CancelTransfer();
          }
        return;
      }
    while (Length > 0)
      {
        if (ReceiveBufPtr >= MaxTotalPacketSize)
          {
            /* too big of a packet, discard everything */
            APRINT(("Huge packet discarded\n"));
            return;
          }
        ReceiveBuffer[ReceiveBufPtr] = *Data;
        if (*Data == 0x0d)
          {
            /* packet terminator, time to process packet */
            HandlePacket(ReceiveBuffer,ReceiveBufPtr);
            ReceiveBufPtr = 0;
          }
         else
          {
            ReceiveBufPtr += 1;
          }
        Data += 1;
        Length -= 1;
      }
  }


void      HandlePacket(char* Data, short Length)
  {
    TransRec    FakePacket;
    short       LengthTemp;

    /* if we don't know how many data bits are being used in the transmission */
    /* path, we try to figure it out here by trying various decodings. */
    APRINT(("+HandlePacket\n"));
    if (NumDataBits == 0)
      {
        char      Temp[MaxTotalPacketSize];

        LengthTemp = Length;
        /* this happens if we don't know what data mode the other side */
        /* is using */
        memcpy(Temp,Data,Length); /* temporarily save packet */
        if (!TransmitDecode(Data,&LengthTemp,8))
          {
            /* ok, 8 bits didn't work. */
           SevenBitPoint:
            LengthTemp = Length;
            memcpy(Data,Temp,Length);
            if (!TransmitDecode(Data,&LengthTemp,7))
              {
                /* this didn't work either.  We'll just have to */
                /* wait for the next packet */
                return;
              }
             else
              {
                /* this did work.  Now we'll try error decoding */
                if (!ErrorDecode(Data,&LengthTemp))
                  {
                    /* Nope, error decoding didn't work */
                    return;
                  }
                 else
                  {
                    NumDataBits = 7;
                    goto StepPoint; /* worked, so we'll use it */
                  }
              }
          }
         else
          {
            /* 8 bits did work.  Now we'll try error decoding */
            if (!ErrorDecode(Data,&LengthTemp))
              {
                /* Nope, we'll try it with 7 bits */
                goto SevenBitPoint;
              }
             else
              {
                /* Yup, 8 did work */
                NumDataBits = 8;
                goto StepPoint;
              }
          }
      }

    if (!TransmitDecode(Data,&Length,NumDataBits))
      {
        APRINT(("-HandlePacket transmit decode failed\n"));
        return; /* packet was corrupted */
      }
    if (!ErrorDecode(Data,&Length))
      {
        APRINT(("-HandlePacket error decode failed\n"));
        return; /* corrupted */
      }
   StepPoint:
    switch (State)
      {
        case ReceiveDoingNothing: /* nothing interesting going on */
          break;
        case ReceiveWaitingForHeader: /* waiting for file header to be sent (we'll ack when we get it) */
         ReceiveHeaderPoint:
          if ((Data[0] == HeaderBlockWord) && (Length == HeaderBlockLength))
            {
              char        FileName[33];
              short       FileRef;
              short       Scan;
              long        OtherScan;
              long        FileAllocateScan;
              char        StupidBuffer[1024];
              short       BytesToWrite;

              APRINT(("Received header, creating file\n"));
              LastTimeOut = TickCount() - RetryIntersticeDelay;
              memcpy(FileName,&(Data[HeaderBlockFileNamePlace + 1]),32);
              if ((unsigned char)Data[HeaderBlockFileNamePlace] > 32)
                {
                  /* truncating improperly constructed name */
                  FileName[32] = 0;
                }
               else
                {
                  FileName[Data[HeaderBlockFileNamePlace]] = 0;
                }
              for (Scan = 0; FileName[Scan] != 0; Scan += 1)
                {
                  /* removing illegal characters from filename */
                  if ((FileName[Scan] < 32) || ((unsigned char)FileName[Scan] > 126))
                    {
                      FileName[Scan] = 'x';
                    }
                  if (FileName[Scan] == '/')
                    {
                      FileName[Scan] = 'x';
                    }
                }
              if (FileName[0] == '.')
                {
                  FileName[0] = 'x'; /* no dot's as first filename char */
                }
             Loop:
              if ((FileRef = open(FileName,O_WRONLY | O_CREAT | O_EXCL,
                S_IRUSR | S_IWUSR)) == -1)
                {
                  if (errno != EEXIST)
                    {
                      APRINT(("creating file error %d\n",errno));
                      CancelTransfer();
                      return;
                    }
                  if (strlen(FileName) < 31)
                    {
                      FileName[strlen(FileName)] = 'x';
                      FileName[strlen(FileName)+1] = 0;
                      goto Loop;
                    }
                   else
                    {
                      short     I;

                      if ((FileName[31] < '0') || (FileName[31] > '9'))
                        {
                          for (I = 0; I < 8; I += 1)
                            {
                              FileName[I + (31 - 7)] = '0';
                            }
                          goto Loop;
                        }
                       else
                        {
                          for (I = 7; I >= 0; I -= 1)
                            {
                              FileName[I + (31 - 7)] += 1;
                              if (FileName[I + (31 - 7)] <= '9')
                                {
                                  goto Loop;
                                }
                               else
                                {
                                  FileName[I + (31 - 7)] = '0';
                                }
                            }
                        }
                      /* if all filenames are exhausted (impossible) we spontaneously */
                      /* splatter blood all over the file system */
                      CancelTransfer();
                      return;
                    }
                }
              CurrentFileDescriptor = FileRef;
              CurrentFileLength = (unsigned char)Data[HeaderBlockFileLengthPlace + 0]
                + ((long)(unsigned char)Data[HeaderBlockFileLengthPlace + 1] << 8)
                + ((long)(unsigned char)Data[HeaderBlockFileLengthPlace + 2] << 16)
                + ((long)(unsigned char)Data[HeaderBlockFileLengthPlace + 3] << 24);
              PreviousBlockIndex = 0;
              /* now we need to allocate all space for the file */
              for (BytesToWrite = 0; BytesToWrite < 1024; BytesToWrite += 1)
                {
                  StupidBuffer[BytesToWrite] = 0;
                }
              for (FileAllocateScan = 0; FileAllocateScan < CurrentFileLength;
                FileAllocateScan += 1024)
                {
                  if (CurrentFileLength - FileAllocateScan >= 1024)
                    {
                      BytesToWrite = 1024;
                    }
                   else
                    {
                      BytesToWrite = CurrentFileLength - FileAllocateScan;
                    }
                  if (write(FileRef,StupidBuffer,BytesToWrite) != BytesToWrite)
                    {
                      /* can't write file data */
                      APRINT(("Failed to write data to file\n"));
                      CancelTransfer();
                      return;
                    }
                }
              while (NumElements(OtherBadBlockList) != 0)
                {
                  ResetScan(OtherBadBlockList);
                  GetNext(OtherBadBlockList,&OtherScan);
                  KillElement(OtherBadBlockList,&OtherScan);
                }
              while (NumElements(BadBlockList) != 0)
                {
                  /* removing any left over entries from bad blocks list */
                  /* this really shouldn't be needed, but just in case */
                  ResetScan(BadBlockList);
                  GetNext(BadBlockList,&OtherScan);
                  KillElement(BadBlockList,&OtherScan);
                }
              for (OtherScan = 0; OtherScan < CurrentFileLength;
                OtherScan += FileXferPacketDataSize)
                {
                  /* start out pretending all blocks are bad */
                  PushElement(BadBlockList,&OtherScan);
                }
              State = WaitingForHeader_Acking;
            }
          break;
        case WaitingForHeader_Acking: /* acking, waiting for first data block */
          if (Data[0] == DataBlockWord)
            {
              APRINT(("WaitingForHeader_Acking, got first data block, changing state\n"));
              State = ReceivingData;
              goto ReceiveDataPoint;
            }
          break;
        case ReceivingData: /* receiving data, waiting for "all done with file", request resends */
          if (Data[0] == DataBlockWord)
            {
              long      ThisBlockIndex; /* byte index */

             ReceiveDataPoint:
              APRINT(("Receiving data... "));
              ThisBlockIndex = (((long)(unsigned char)Data[DataBlock1KIndex + 0])
                + ((long)(unsigned char)Data[DataBlock1KIndex + 1] << 8)
                + ((long)(unsigned char)Data[DataBlock1KIndex + 2] << 16))
                * FileXferPacketDataSize;
              APRINT(("Packet %d (%d)\n",ThisBlockIndex/FileXferPacketDataSize,ThisBlockIndex));
              Length = Length - DataBlockLength; /* remove overhead to find total bytes */
              if (((Length < FileXferPacketDataSize) && (CurrentFileLength
                - ThisBlockIndex > FileXferPacketDataSize))
                || (Length + ThisBlockIndex > CurrentFileLength))
                {
                  /* this means packet is too short or too long */
                  /* all packets except last MUST be the full length.  Also, */
                  /* the last can't exceed the end of file. */
                  APRINT(("Packet wrong size, discarding\n"));
                  return; /* discard it */
                }
              if (State == ReceivingData)
                {
                  /* this loop adds packets to the bad list that are missing */
                  /* between the previously received one and the current one */
                  /* we only do this during the streaming phase */
                  while (ThisBlockIndex - FileXferPacketDataSize > PreviousBlockIndex)
                    {
                      PreviousBlockIndex += FileXferPacketDataSize;
                      PushElement(OtherBadBlockList,&PreviousBlockIndex);
                      APRINT(("Missing block %d\n",(int)(PreviousBlockIndex/FileXferPacketDataSize)));
                    }
                  if (ThisBlockIndex > PreviousBlockIndex)
                    {
                      PreviousBlockIndex = ThisBlockIndex;
                    }
                }
              if (State == ReceivingData_Cleanup)
                {
                  long      TempBadBlock;

                  /* here we check to see if the packet we just received was */
                  /* the one we just requested to be resent.  If NOT, then we */
                  /* assume the remote buffer is full and reset our timeout */
                  ResetScan(BadBlockList);
                  GetNext(BadBlockList,&TempBadBlock);
                  if (TempBadBlock == ThisBlockIndex)
                    {
                      /* block we received was probably an answer to our request */
                      LastTimeOut = TickCount() - RetryIntersticeDelay;
                    }
                   else
                    {
                      /* block we received was probably due to a full buffer, */
                      /* so we will delay our request for a bad block. */
                      LastTimeOut = TickCount();
                      APRINT(("Looks like remote buffer full; resetting timeout counter\n"));
                    }
                }
              if (KillElement(BadBlockList,&ThisBlockIndex))
                {
                  /* if kill was successful, then we haven't received block yet */
                  APRINT(("Saving block\n"));
                  if (PreviousBlockIndex < ThisBlockIndex)
                    {
                      PreviousBlockIndex = ThisBlockIndex;
                    }
                  lseek(CurrentFileDescriptor,ThisBlockIndex,SEEK_SET);
                  write(CurrentFileDescriptor,&(Data[DataBlockDataStart]),Length);
                }
               else
                {
                  APRINT(("BLOCK WASN'T SAVED!!!\n"));
                }
            }
          if ((Data[0] == DoneSendingResolveErrorsWord)
            && (Length == DoneSendingResolveErrorsLength))
            {
              APRINT(("Sender is done, switching to cleanup mode\n"));
              /* this point means the sender is done sending, time for error correction */
              State = ReceivingData_Cleanup;
              /* now we reset the timeout so that bad packet requests get send */
              /* RIGHT AWAY, not after the timeout period */
              LastTimeOut = TickCount() - RetryIntersticeDelay;
            }
          break;
        case ReceivingData_Cleanup: /* "all done with file" gotten, correcting errors */
          if (Data[0] == DataBlockWord)
            {
              goto ReceiveDataPoint; /* handle this resent packet */
            }
          break;
        case SendGoodWaitForHeader: /* sending "all done", waiting for next header or quit */
          if ((Data[0] == HeaderBlockWord) && (Length == HeaderBlockLength))
            {
              APRINT(("SendGoodWaitForHeader: got header, changing state\n"));
              State = ReceiveWaitingForHeader;
              goto ReceiveHeaderPoint;
            }
          break;
        case ReceiveGoAway:
          APRINT(("ReceiveGoAway; don't know what to do\n"));
          /* nothing to do here */
          break;
        default:
          APRINT(("unknown state encountered\n"));
          exit(1);
      }
    APRINT(("-HandlePacket\n"));
  }


void      CancelTransfer(void)
  {
    short     Scan;

    APRINT(("+CancelTransfer\n"));
    for (Scan = 0; Scan < ABORTTHRESHHOLD + 3; Scan += 1)
      {
        SendBuffer[Scan] = CANCELCHARACTER;
      }
    SendBufPtr = ABORTTHRESHHOLD + 3;
    State = ReceiveGoAway; /* go away as soon as data is sent */
    APRINT(("-CancelTransfer\n"));
  }
