/* SendFile.c */

#include "commonstuff.h"

enum
  {
    SendDoingNothing, /* before anything starts */
    SendingHeader, /* sending header, waiting for receiver to acknowledge */
    SendingData, /* sending data blocks, receiving resend requests */
    SendingWantErrors, /* done sending data, waiting for error requests */
    SendFixingErrors, /* waiting for receiver to say "ok, all good" & correcting errs */
    SendGoAway /* as soon as send buffer flushed, die */
  };


short           ReceiveBufPtr;
short           SendBufPtr;
char**          ListOfFiles; /* from argv */
int             NumFiles; /* from argc */
short           CurrentFileDescriptor;
long            CurrentFileIndex;
short           State;
unsigned long   LastTimeOut;
CSack*          ListOfKnownBadPackets;
short           NumDataBits;
char            ReceiveBuffer[MaxTotalPacketSize];
char            SendBuffer[MaxTotalPacketSize];

void      InitModule(int argc, char* argv[]);
void      DoIdle(void);
void      ReceiveBytes(char* Data, long Length);
void      HandlePacket(char* Data, short Length);
void      CancelTransfer(void);


void      InitModule(int argc, char* argv[])
  {
    APRINT(("+InitModule\n"));

    write(0,"\x18Majik",6); /* send "invite to download" command */
    sleep(2); /* pause */

    ReceiveBufPtr = 0;
    SendBufPtr = 0;
    ListOfKnownBadPackets = NewCSack();
    NumDataBits = 0;

    if ((argc > 0) && (strcmp(argv[0],"-7") == 0))
      {
        NumDataBits = 7;
        argc -= 1;
        argv += 1;
      }
     else
      {
        NumDataBits = 8;
      }

    State = SendingHeader;
    ListOfFiles = argv;
    NumFiles = argc;
    LastTimeOut = TickCount() - RetryIntersticeDelay;

    APRINT(("-InitModule\n"));
  }


#define LocalBufSize (255)

void      DoIdle(void)
  {
    TransRec      FakePacket;
    long          FileLength;
    long          FileBytes;
    char          LocalReadBuffer[LocalBufSize];

   TryAgainPoint:
    /* first, see if we have received any data */
    FileBytes = LocalBufSize;
    FileBytes = read(0,LocalReadBuffer,FileBytes);
    if (FileBytes < 0)
      {
        FileBytes = 0;
      }
    if (FileBytes > 0)
      {
        ReceiveBytes(LocalReadBuffer,FileBytes);
        goto TryAgainPoint;
      }

    /* now, flush our internal send buffer */
    if (SendBufPtr != 0)
      {
        long        NumBytesWritten;

        NumBytesWritten = write(1,SendBuffer,SendBufPtr);
        if (NumBytesWritten <= 0)
          {
            return; /* failed altogether */
          }
        if (NumBytesWritten < SendBufPtr)
          {
          	long        Scan;

            SendBufPtr -= NumBytesWritten;
            for (Scan = 0; Scan < SendBufPtr; Scan += 1)
              {
                SendBuffer[Scan] = SendBuffer[Scan + NumBytesWritten];
              }
            return;
          }
         else
          {
            SendBufPtr = 0;
          }
      }

    /* now, we enter the state machine to see what's going on. */
    switch (State)
      {
        case SendDoingNothing: /* before anything starts */
          break;
        case SendingHeader: /* sending header, waiting for receiver to acknowledge */
          if (TickCount() - LastTimeOut >= RetryIntersticeDelay)
            {
              int         TempFileDesc;

              APRINT(("Sending header\n"));
              if (NumFiles == 0)
                {
                  short     Scan;

                  /* no more files, so cancelling transfer */
                  APRINT(("Sending header: no more files, cancelling\n"));
                  for (Scan = 0; Scan < ABORTTHRESHHOLD + 3; Scan += 1)
                    {
                      SendBuffer[Scan] = CANCELCHARACTER;
                    }
                  SendBufPtr = ABORTTHRESHHOLD + 3;
                  State = SendGoAway;
                  goto TryAgainPoint;
                }
              LastTimeOut = TickCount();
              FakePacket.Data = SendBuffer;
              FakePacket.Length = HeaderBlockLength;
              SendBuffer[0] = HeaderBlockWord;
              strncpy(&(SendBuffer[HeaderBlockFileNamePlace + 1]),ListOfFiles[0],32);
              if (strlen(ListOfFiles[0]) > 32)
                {
                  SendBuffer[HeaderBlockFileNamePlace] = 32;
                }
               else
                {
                  SendBuffer[HeaderBlockFileNamePlace] = strlen(ListOfFiles[0]);
                }
              TempFileDesc = open(ListOfFiles[0],O_RDONLY,0);
              if (TempFileDesc == -1)
                {
                  /* if an error occurred, then skip to the next file */
                  NumFiles -= 1;
                  ListOfFiles += 1;
                  goto TryAgainPoint;
                }
              {
                struct stat     Info;

                if (-1 == fstat(TempFileDesc,&Info))
                  {
                    APRINT(("Sending header:  couldn't fstat, aborting\n"));
                    CancelTransfer();
                    return;
                  }
                FileLength = Info.st_size;
              }
              close(TempFileDesc);
              SendBuffer[HeaderBlockFileLengthPlace + 0] = FileLength & 0x00ff;
              SendBuffer[HeaderBlockFileLengthPlace + 1] = (FileLength >> 8) & 0x00ff;
              SendBuffer[HeaderBlockFileLengthPlace + 2] = (FileLength >> 16) & 0x00ff;
              SendBuffer[HeaderBlockFileLengthPlace + 3] = (FileLength >> 24) & 0x00ff;
              ErrorEncode(&FakePacket);
              TransmitEncode(&FakePacket,NumDataBits);
              SendBufPtr = FakePacket.Length;
              goto TryAgainPoint;
            }
          break;
        case SendingData: /* sending data blocks, receiving resend requests */
          {
            long      PacketIndex;
            long      DataSize;
            char      Number[3];
            short     Percent;

           ResendAPacketPoint:
              {
                struct stat     Info;

                APRINT(("Sending data\n"));
                if (-1 == fstat(CurrentFileDescriptor,&Info))
                  {
                    APRINT(("Resending data: couldn't fstat, aborting\n"));
                    CancelTransfer();
                    return;
                  }
                FileLength = Info.st_size;
              }
            if (NumElements(ListOfKnownBadPackets) != 0)
              {
                APRINT(("Resending:  there are known bad packets... "));
                ResetScan(ListOfKnownBadPackets);
                GetNext(ListOfKnownBadPackets,&PacketIndex);
                KillElement(ListOfKnownBadPackets,&PacketIndex);
                APRINT(("Resending packet indexed by %d\n",PacketIndex));
              }
             else
              {
                if (CurrentFileIndex >= FileLength)
                  {
                    /* this means we have finished with the file, so */
                    /* we can't send any more packets until one is requested */
                    APRINT(("End of file, switching to SendingWantErrors mode\n"));
                    State = SendingWantErrors;
                    LastTimeOut = TickCount() - RetryIntersticeDelay;
                    return;
                  }
                PacketIndex = CurrentFileIndex;
                CurrentFileIndex += FileXferPacketDataSize;
              }
            /* figuring out how big to make packet */
            if (FileLength - PacketIndex < FileXferPacketDataSize)
              {
                DataSize = FileLength - PacketIndex;
              }
             else
              {
                DataSize = FileXferPacketDataSize;
              }
            /* making packet */
            FakePacket.Data = SendBuffer;
            FakePacket.Length = DataBlockLength + DataSize;
            SendBuffer[0] = DataBlockWord;
            SendBuffer[DataBlock1KIndex + 0] = ((PacketIndex
              / FileXferPacketDataSize)) & 0x00ff;
            SendBuffer[DataBlock1KIndex + 1] = ((PacketIndex
              / FileXferPacketDataSize) >> 8) & 0x00ff;
            SendBuffer[DataBlock1KIndex + 2] = ((PacketIndex
              / FileXferPacketDataSize) >> 16) & 0x00ff;
            APRINT(("Packet ID number %d\n",PacketIndex/FileXferPacketDataSize));
            lseek(CurrentFileDescriptor,PacketIndex,SEEK_SET);
            if (DataSize != read(CurrentFileDescriptor,&(SendBuffer[DataBlockDataStart]),DataSize))
              {
                APRINT(("Couldn't read packet data from file, aborting\n"));
                CancelTransfer();
                return;
              }
            ErrorEncode(&FakePacket);
            TransmitEncode(&FakePacket,NumDataBits);
            SendBufPtr = FakePacket.Length;
            goto TryAgainPoint;
          }
          break;
        case SendingWantErrors:
          if (TickCount() - LastTimeOut >= RetryIntersticeDelay)
            {
              APRINT(("Sending 'want errors'\n"));
              LastTimeOut = TickCount();
              FakePacket.Data = SendBuffer;
              FakePacket.Length = DoneSendingResolveErrorsLength;
              SendBuffer[0] = DoneSendingResolveErrorsWord;
              ErrorEncode(&FakePacket);
              TransmitEncode(&FakePacket,NumDataBits);
              SendBufPtr = FakePacket.Length;
              goto TryAgainPoint;
            }
          break;
        case SendFixingErrors: /* waiting for receiver to say "ok, all good" & correcting errs */
          /* at this point, we can only resend in response to an error. */
          /* we receive errors in HandlePacket, which get stuck in ListOfKnownBadPackets */
          /* for us.  So we use the normal handling method to resend them */
          if (NumElements(ListOfKnownBadPackets) != 0)
            {
              APRINT(("SendFixingErrors: there are known bad packets\n"));
              goto ResendAPacketPoint;
            }
          break;
        case SendGoAway:
          APRINT(("SendGoAway:  going away\n"));
          QuitFlag = True;
          return;
        default:
          APRINT(("Error, unknown state in idle loop, exiting\n"));
          exit(1);
      }
  }


void      ReceiveBytes(char* Data, long Length)
  {
    short       AccumulatedAborts;
    short       MaxAborts;
    short       DataScan;

    AccumulatedAborts = 0;
    MaxAborts = 0;
    for (DataScan = 0; DataScan < ReceiveBufPtr; DataScan += 1)
      {
        if (ReceiveBuffer[DataScan] == CANCELCHARACTER)
          {
            AccumulatedAborts += 1;
          }
         else
          {
            AccumulatedAborts = 0;
          }
        if (AccumulatedAborts > MaxAborts)
          {
            MaxAborts = AccumulatedAborts;
          }
      }
    for (DataScan = 0; DataScan < Length; DataScan += 1)
      {
        if (Data[DataScan] == CANCELCHARACTER)
          {
            AccumulatedAborts += 1;
          }
         else
          {
            AccumulatedAborts = 0;
          }
        if (AccumulatedAborts > MaxAborts)
          {
            MaxAborts = AccumulatedAborts;
          }
      }
    if (MaxAborts > ABORTTHRESHHOLD)
      {
        APRINT(("Receiving data:  abort sequence received.\n"));
        CancelTransfer();
        return;
      }
    while (Length > 0)
      {
        if (ReceiveBufPtr >= MaxTotalPacketSize)
          {
            /* too big of a packet, discard everything */
            APRINT(("Huge packet, discarded\n"));
            return;
          }
        ReceiveBuffer[ReceiveBufPtr] = *Data;
        if (*Data == 0x0d)
          {
            /* packet terminator, time to process packet */
            HandlePacket(ReceiveBuffer,ReceiveBufPtr);
            ReceiveBufPtr = 0;
          }
         else
          {
            ReceiveBufPtr += 1;
          }
        Data += 1;
        Length -= 1;
      }
  }


void      HandlePacket(char* Data, short Length)
  {
    TransRec    FakePacket;

    APRINT(("+HandlePacket\n"));
    if (!TransmitDecode(Data,&Length,NumDataBits))
      {
        APRINT(("-HandlePacket transmit decode failed\n"));
        return; /* packet was corrupted */
      }
    if (!ErrorDecode(Data,&Length))
      {
        APRINT(("-HandlePacket error decode failed\n"));
        return; /* corrupted */
      }
    switch (State)
      {
        case SendDoingNothing: /* before anything starts */
          break;
        case SendingHeader: /* sending header, waiting for receiver to acknowledge */
          if ((Data[0] == GotHeaderStartSendingWord)
            && (Length == GotHeaderStartSendingLength))
            {
              short       TempFileDesc;

              APRINT(("Sending Header: got acknowledge\n"));
              if (NumFiles != 0)
                {
                  LastTimeOut = TickCount() - RetryIntersticeDelay;
                  State = SendingData;
                  SendBuffer[0] = HeaderBlockWord;
                  TempFileDesc = open(ListOfFiles[0],O_RDONLY,0);
                  if (TempFileDesc == -1)
                    {
                      APRINT(("Couldn't open file, aborting\n"));
                      CancelTransfer();
                      return;
                    }
                  CurrentFileDescriptor = TempFileDesc;
                  CurrentFileIndex = 0;
                  NumFiles -= 1;
                  ListOfFiles += 1;
                }
               else
                {
                  short     Scan;

                  APRINT(("No files to send, quitting\n"));
                  for (Scan = 0; Scan < ABORTTHRESHHOLD + 3; Scan += 1)
                    {
                      SendBuffer[Scan] = CANCELCHARACTER;
                    }
                  SendBufPtr = ABORTTHRESHHOLD + 3;
                  State = SendGoAway;
                }
            }
          break;
        case SendingData: /* sending data blocks, receiving resend requests */
         ResendDataPoint:
          if ((Data[0] == ResendRequestWord) && (Length == ResendRequestLength))
            {
              long      ByteIndex;

              APRINT(("Sending data, got a resend request... \n"));
              ByteIndex = ((unsigned char)Data[ResendRequest1KIndex + 0]
                + ((unsigned long)(unsigned char)Data[ResendRequest1KIndex + 1] << 8)
                + ((unsigned long)(unsigned char)Data[ResendRequest1KIndex + 2] << 16))
                * FileXferPacketDataSize;
              APRINT(("Packet index %d\n",ByteIndex));
              /* before we save a packet to be resent, we remove it from */
              /* the list so there are not multiple copies in the list. */
              KillElement(ListOfKnownBadPackets,&ByteIndex);
              PushElement(ListOfKnownBadPackets,&ByteIndex);
              APRINT(("BadBlock %d\n",(int)(ByteIndex / FileXferPacketDataSize)));
            }
          break;
        case SendingWantErrors:
          if (((Data[0] == GoodGotoNextFileWord) && (Length == GoodGotoNextFileLength))
            || ((Data[0] == ResendRequestWord) && (Length == ResendRequestLength)))
            {
              APRINT(("SendingWantErrors:  ack received, switching modes\n"));
              State = SendFixingErrors;
              goto FixErrorsPoint;
            }
          break;
        case SendFixingErrors: /* waiting for receiver to say "ok, all good" & correcting errs */
         FixErrorsPoint:
          if ((Data[0] == GoodGotoNextFileWord) && (Length == GoodGotoNextFileLength))
            {
              APRINT(("SendFixingErrors: file finished, go to next\n"));
              close(CurrentFileDescriptor);
              LastTimeOut = TickCount() - RetryIntersticeDelay;
              State = SendingHeader; /* send header for next file */
            }
          if ((Data[0] == ResendRequestWord) && (Length == ResendRequestLength))
            {
              APRINT(("SendFixingErrors: resending something\n"));
              goto ResendDataPoint;
            }
          break;
        case SendGoAway:
          APRINT(("SendGoAway: received a packet, but can't deal with it\n"));
          break; /* nothing to do */
        default:
          APRINT(("unknown state encountered\n"));
          exit(1);
      }
    APRINT(("-HandlePacket\n"));
  }


void      CancelTransfer(void)
  {
    short     Scan;

    APRINT(("+CancelTransfer\n"));
    for (Scan = 0; Scan < ABORTTHRESHHOLD + 3; Scan += 1)
      {
        SendBuffer[Scan] = CANCELCHARACTER;
      }
    SendBufPtr = ABORTTHRESHHOLD + 3;
    State = SendGoAway; /* go away as soon as data is sent */
    APRINT(("-CancelTransfer\n"));
  }
